#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import contextlib
import logging
import warnings
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import torch

from ludwig.api_annotations import DeveloperAPI
from ludwig.backend import Backend, LOCAL_BACKEND
from ludwig.constants import (
    BFILL,
    CHECKSUM,
    COLUMN,
    DEFAULTS,
    DROP_ROW,
    ENCODER,
    FFILL,
    FILL_WITH_CONST,
    FILL_WITH_FALSE,
    FILL_WITH_MEAN,
    FILL_WITH_MODE,
    FULL,
    META,
    MIN_DATASET_SPLIT_ROWS,
    MODEL_ECD,
    NAME,
    NUMBER,
    PREPROCESSING,
    PROC_COLUMN,
    SPLIT,
    SRC,
    TEST,
    TRAINING,
    TYPE,
    VALIDATION,
)
from ludwig.data.cache.manager import DatasetCache
from ludwig.data.cache.types import wrap
from ludwig.data.concatenate_datasets import concatenate_df, concatenate_files, concatenate_splits
from ludwig.data.dataset.base import Dataset
from ludwig.data.split import get_splitter, split_dataset
from ludwig.data.utils import set_fixed_split
from ludwig.datasets import load_dataset_uris
from ludwig.features.feature_registries import get_base_type_registry
from ludwig.models.embedder import create_embed_batch_size_evaluator, create_embed_transform_fn
from ludwig.schema.encoders.utils import get_encoder_cls
from ludwig.types import FeatureConfigDict, PreprocessingConfigDict, TrainingSetMetadataDict
from ludwig.utils import data_utils, strings_utils
from ludwig.utils.backward_compatibility import upgrade_metadata
from ludwig.utils.data_utils import (
    CACHEABLE_FORMATS,
    CSV_FORMATS,
    DATA_TEST_PARQUET_FP,
    DATA_TRAIN_HDF5_FP,
    DATA_TRAIN_PARQUET_FP,
    DATA_VALIDATION_PARQUET_FP,
    DATAFRAME_FORMATS,
    DICT_FORMATS,
    EXCEL_FORMATS,
    FEATHER_FORMATS,
    figure_data_format,
    FWF_FORMATS,
    get_split_path,
    HDF5_FORMATS,
    HTML_FORMATS,
    JSON_FORMATS,
    JSONL_FORMATS,
    ORC_FORMATS,
    override_in_memory_flag,
    PARQUET_FORMATS,
    PICKLE_FORMATS,
    read_csv,
    read_excel,
    read_feather,
    read_fwf,
    read_html,
    read_json,
    read_jsonl,
    read_orc,
    read_parquet,
    read_pickle,
    read_sas,
    read_spss,
    read_stata,
    read_tsv,
    SAS_FORMATS,
    SPSS_FORMATS,
    STATA_FORMATS,
    TSV_FORMATS,
)
from ludwig.utils.dataframe_utils import is_dask_series_or_df
from ludwig.utils.defaults import (
    default_prediction_preprocessing_parameters,
    default_random_seed,
    default_training_preprocessing_parameters,
)
from ludwig.utils.fs_utils import file_lock, path_exists
from ludwig.utils.misc_utils import get_from_registry, merge_dict
from ludwig.utils.types import DataFrame, Series

REPARTITIONING_FEATURE_TYPES = {"image", "audio"}

logger = logging.getLogger(__name__)


class DataFormatPreprocessor(ABC):
    @staticmethod
    @abstractmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        pass

    @staticmethod
    @abstractmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        pass

    @staticmethod
    @abstractmethod
    def prepare_processed_data(
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
    ):
        pass


class DictPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        num_overrides = override_in_memory_flag(features, True)
        if num_overrides > 0:
            logger.warning("Using in_memory = False is not supported " "with {} data format.".format("dict"))

        df_engine = backend.df_engine
        if dataset is not None:
            dataset = df_engine.from_pandas(pd.DataFrame(dataset))
        if training_set is not None:
            training_set = df_engine.from_pandas(pd.DataFrame(training_set))
        if validation_set is not None:
            validation_set = df_engine.from_pandas(pd.DataFrame(validation_set))
        if test_set is not None:
            test_set = df_engine.from_pandas(pd.DataFrame(test_set))

        return _preprocess_df_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            training_set_metadata=training_set_metadata,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset, training_set_metadata = build_dataset(
            pd.DataFrame(dataset),
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class DataFramePreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        num_overrides = override_in_memory_flag(features, True)
        if num_overrides > 0:
            logger.warning("Using in_memory = False is not supported " "with {} data format.".format("dataframe"))

        if isinstance(dataset, pd.DataFrame):
            dataset = backend.df_engine.from_pandas(dataset)

        return _preprocess_df_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            training_set_metadata=training_set_metadata,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        if isinstance(dataset, pd.DataFrame):
            dataset = backend.df_engine.from_pandas(dataset)

        dataset, training_set_metadata = build_dataset(
            dataset,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class CSVPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_csv,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_csv(dataset, df_lib=backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class TSVPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_tsv,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_tsv(dataset, df_lib=backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class JSONPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_json,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_json(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class JSONLPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_jsonl,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_jsonl(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class ExcelPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_excel,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_excel(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class ParquetPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_parquet,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_parquet(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None

    @staticmethod
    def prepare_processed_data(
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
    ):
        test_set = test_set if test_set and path_exists(test_set) else None
        if test_set and isinstance(test_set, str) and DATA_TEST_PARQUET_FP not in training_set_metadata:
            training_set_metadata[DATA_TEST_PARQUET_FP] = test_set

        validation_set = validation_set if validation_set and path_exists(validation_set) else None
        if (
            validation_set
            and isinstance(validation_set, str)
            and DATA_VALIDATION_PARQUET_FP not in training_set_metadata
        ):
            training_set_metadata[DATA_VALIDATION_PARQUET_FP] = validation_set

        if training_set and isinstance(training_set, str) and DATA_TRAIN_PARQUET_FP not in training_set_metadata:
            training_set_metadata[DATA_TRAIN_PARQUET_FP] = training_set
        return training_set, test_set, validation_set, training_set_metadata


class PicklePreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_pickle,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_pickle(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class FatherPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_feather,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_feather(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class FWFPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_fwf,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_fwf(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class HTMLPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_html,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_html(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class ORCPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_orc,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_orc(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class SASPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_sas,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_sas(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class SPSSPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_spss,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_spss(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class StataPreprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return _preprocess_file_for_training(
            config,
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            read_fn=read_stata,
            training_set_metadata=training_set_metadata,
            skip_save_processed_input=skip_save_processed_input,
            preprocessing_params=preprocessing_params,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        dataset_df = read_stata(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset
        dataset, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="prediction",
            metadata=training_set_metadata,
            backend=backend,
            callbacks=callbacks,
        )
        return dataset, training_set_metadata, None


class HDF5Preprocessor(DataFormatPreprocessor):
    @staticmethod
    def preprocess_for_training(
        config,
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
        callbacks=None,
    ):
        return HDF5Preprocessor.prepare_processed_data(
            features,
            dataset,
            training_set,
            validation_set,
            test_set,
            training_set_metadata,
            skip_save_processed_input,
            preprocessing_params,
            backend,
            random_seed,
        )

    @staticmethod
    def preprocess_for_prediction(dataset, features, preprocessing_params, training_set_metadata, backend, callbacks):
        hdf5_fp = dataset
        dataset = load_hdf5(dataset, preprocessing_params, backend, split_data=False, shuffle_training=False)
        return dataset, training_set_metadata, hdf5_fp

    @staticmethod
    def prepare_processed_data(
        features,
        dataset=None,
        training_set=None,
        validation_set=None,
        test_set=None,
        training_set_metadata=None,
        skip_save_processed_input=False,
        preprocessing_params=default_training_preprocessing_parameters,
        backend=LOCAL_BACKEND,
        random_seed=default_random_seed,
    ):
        if dataset is None and training_set is None:
            raise ValueError("One of `dataset` or `training_set` must be not None")
        not_none_set = dataset if dataset is not None else training_set

        if not training_set_metadata:
            raise ValueError("When providing HDF5 data, " "training_set_metadata must not be None.")

        logger.info("Using full hdf5 and json")

        if DATA_TRAIN_HDF5_FP not in training_set_metadata:
            logger.warning(
                "data_train_hdf5_fp not present in training_set_metadata. "
                "Adding it with the current HDF5 file path {}".format(not_none_set)
            )
            training_set_metadata[DATA_TRAIN_HDF5_FP] = not_none_set

        elif training_set_metadata[DATA_TRAIN_HDF5_FP] != not_none_set:
            logger.warning(
                "data_train_hdf5_fp in training_set_metadata is {}, "
                "different from the current HDF5 file path {}. "
                "Replacing it".format(training_set_metadata[DATA_TRAIN_HDF5_FP], not_none_set)
            )
            training_set_metadata[DATA_TRAIN_HDF5_FP] = not_none_set

        if dataset is not None:
            training_set, test_set, validation_set = load_hdf5(
                dataset, preprocessing_params, backend, shuffle_training=True
            )

        elif training_set is not None:
            kwargs = dict(preprocessing_params=preprocessing_params, backend=backend, split_data=False)
            training_set = load_hdf5(training_set, shuffle_training=True, **kwargs)

            if validation_set is not None:
                validation_set = load_hdf5(validation_set, shuffle_training=False, **kwargs)

            if test_set is not None:
                test_set = load_hdf5(test_set, shuffle_training=False, **kwargs)

        return training_set, test_set, validation_set, training_set_metadata


data_format_preprocessor_registry = {
    **{fmt: DictPreprocessor for fmt in DICT_FORMATS},
    **{fmt: DataFramePreprocessor for fmt in DATAFRAME_FORMATS},
    **{fmt: CSVPreprocessor for fmt in CSV_FORMATS},
    **{fmt: TSVPreprocessor for fmt in TSV_FORMATS},
    **{fmt: JSONPreprocessor for fmt in JSON_FORMATS},
    **{fmt: JSONLPreprocessor for fmt in JSONL_FORMATS},
    **{fmt: ExcelPreprocessor for fmt in EXCEL_FORMATS},
    **{fmt: ParquetPreprocessor for fmt in PARQUET_FORMATS},
    **{fmt: PicklePreprocessor for fmt in PICKLE_FORMATS},
    **{fmt: FWFPreprocessor for fmt in FWF_FORMATS},
    **{fmt: FatherPreprocessor for fmt in FEATHER_FORMATS},
    **{fmt: HTMLPreprocessor for fmt in HTML_FORMATS},
    **{fmt: ORCPreprocessor for fmt in ORC_FORMATS},
    **{fmt: SASPreprocessor for fmt in SAS_FORMATS},
    **{fmt: SPSSPreprocessor for fmt in SPSS_FORMATS},
    **{fmt: StataPreprocessor for fmt in STATA_FORMATS},
    **{fmt: HDF5Preprocessor for fmt in HDF5_FORMATS},
}


def build_dataset(
    dataset_df,
    features,
    global_preprocessing_parameters,
    mode,
    metadata=None,
    backend=LOCAL_BACKEND,
    random_seed=default_random_seed,
    skip_save_processed_input=False,
    callbacks=None,
):
    """Builds a dataset from a dataframe and a list of features.

    Args:
        dataset_df: Pandas or Dask dataframe
        features: List of features
        global_preprocessing_parameters: Global preprocessing parameters
        mode: One of ['training', 'prediction']
        metadata: Training set metadata if available
        backend: Backend
        random_seed: Random seed
        skip_save_processed_input: Whether to skip saving the processed input
        callbacks: List of callbacks

    Returns:
        A tuple of (dataset, metadata)
    """

    df_engine = backend.df_engine

    if df_engine.partitioned:
        if any(f["type"] in REPARTITIONING_FEATURE_TYPES for f in features) and dataset_df.npartitions > 1:
            # A globally unique index only matters if you know that there will be a repartition downstream for some
            # particular feature, i.e. for Image and Audio features on a Ray backend.
            # - There is a join operation in `df_like`, and the only way to do the operation is if the partitions across
            #   all feature columns are aligned.
            # - In order to align the partitions, we require a way of matching samples to one another across all
            #   partitions. Therefore, we must reset_index to create a globally unique index.
            # - If the number of partitions is 1, it is *highly likely* the index is globally unique. Auto-assigned
            #   Dask indices in this case are unique, and we pd.concat train, val, and test sets with ignore_index=True
            # If there will NOT be a repartition downstream, then we can skip this step.
            # - In this case, the partitions should remain aligned throughout.
            # - Further, while the indices might not be globally unique, they should be unique within each partition.
            # - These two properties make it possible to do the join op within each partition without a global index.
            logger.warning(
                f"Dataset has {dataset_df.npartitions} partitions and feature types that cause repartitioning. "
                f"Resetting index to ensure globally unique indices."
            )
            dataset_df = df_engine.reset_index(dataset_df)

    dataset_df = df_engine.parallelize(dataset_df)

    if mode == "training":
        sample_ratio = global_preprocessing_parameters["sample_ratio"]
        if sample_ratio < 1.0:
            if not df_engine.partitioned and len(dataset_df) * sample_ratio < 1:
                raise ValueError(
                    f"sample_ratio {sample_ratio} is too small for dataset of length {len(dataset_df)}. "
                    f"Please increase sample_ratio or use a larger dataset."
                )

            logger.debug(f"sample {sample_ratio} of data")
            dataset_df = dataset_df.sample(frac=sample_ratio, random_state=random_seed)

    # If persisting DataFrames in memory is enabled, we want to do this after
    # each batch of parallel ops in order to avoid redundant computation
    dataset_df = df_engine.persist(dataset_df)

    if mode == "training":
        default_preprocessing_parameters = default_training_preprocessing_parameters
    elif mode == "prediction":
        default_preprocessing_parameters = default_prediction_preprocessing_parameters
    else:
        raise ValueError(f"Invalid mode {mode}")
    global_preprocessing_parameters = merge_dict(default_preprocessing_parameters, global_preprocessing_parameters)

    # Get all the unique preprocessing features to compute
    feature_configs = []
    feature_hashes = set()
    for feature in features:
        if feature[PROC_COLUMN] not in feature_hashes:
            feature_configs.append(feature)
            feature_hashes.add(feature[PROC_COLUMN])

    dataset_cols = {}
    for feature_config in feature_configs:
        dataset_cols[feature_config[COLUMN]] = dataset_df[feature_config[COLUMN]]

    logger.debug("build preprocessing parameters")
    feature_name_to_preprocessing_parameters = build_preprocessing_parameters(
        dataset_cols, feature_configs, global_preprocessing_parameters, backend, metadata=metadata
    )

    # Happens after preprocessing parameters are built, so we can use precomputed fill values.
    logger.debug("handle missing values")

    # In some cases, there can be a (temporary) mismatch between the dtype of the column and the type expected by the
    # preprocessing config (e.g., a categorical feature represented as an int-like column). In particular, Dask
    # may raise an error even when there are no missing values in the column itself.
    #
    # Since we immediately cast all columns in accordance with their expected feature types after filling missing
    # values, we work around the above issue by temporarily treating all columns as object dtype.
    for col_key in dataset_cols:
        dataset_cols[col_key] = dataset_cols[col_key].astype(object)

    for feature_config in feature_configs:
        preprocessing_parameters = feature_name_to_preprocessing_parameters[feature_config[NAME]]
        handle_missing_values(dataset_cols, feature_config, preprocessing_parameters, backend)

    # Happens after missing values are handled to avoid NaN casting issues.
    logger.debug("cast columns")
    cast_columns(dataset_cols, feature_configs, backend)

    for callback in callbacks or []:
        callback.on_build_metadata_start(dataset_df, mode)

    logger.debug("build metadata")
    metadata = build_metadata(
        metadata, feature_name_to_preprocessing_parameters, dataset_cols, feature_configs, backend
    )

    for callback in callbacks or []:
        callback.on_build_metadata_end(dataset_df, mode)

    for callback in callbacks or []:
        callback.on_build_data_start(dataset_df, mode)

    logger.debug("build data")
    proc_cols = build_data(dataset_cols, feature_configs, metadata, backend, skip_save_processed_input)

    for callback in callbacks or []:
        callback.on_build_data_end(dataset_df, mode)

    # Get any additional columns needed for splitting downstream, otherwise they will not be
    # included in the preprocessed output.
    split_params = global_preprocessing_parameters.get(SPLIT, {})
    if "type" not in split_params and SPLIT in dataset_df:
        warnings.warn(
            'Detected "split" column in the data, but using default split type '
            '"random". Did you mean to set split type to "fixed"?'
        )

    splitter = get_splitter(**split_params)
    for column in splitter.required_columns:
        if column not in dataset_df:
            warnings.warn(
                f"column: '{column}' is required by the dataset splitter with params: {split_params}, but '{column}' "
                f"is not present in the `dataset_df` with columns: {dataset_df.columns}. This is acceptable in a "
                "serving setting where dataset splitting is irrelevant. You may see this warning if, for example, the "
                "model was trained with a configuration that used a stratified split on the target column, but for "
                "live predictions, a value for the target column is not to be provided."
            )
            continue
        proc_cols[column] = dataset_df[column]

    # TODO pyarrow: this is needed for caching to work with pyarrow. if removed, the following error is raised:
    # "pyarrow.lib.ArrowInvalid: Can only convert 1-dimensional array values". The data is reshaped when loaded
    # by the batcher in the RayDataset class (see _prepare_batch).
    if not skip_save_processed_input and backend.cache.data_format == "parquet":
        for feature in features:
            name = feature[NAME]
            proc_column = feature[PROC_COLUMN]
            reshape = metadata[name].get("reshape")
            if reshape is not None:
                proc_cols[proc_column] = backend.df_engine.map_objects(proc_cols[proc_column], lambda x: x.reshape(-1))

    # Implements an outer join of proc_cols
    dataset = backend.df_engine.df_like(dataset_df, proc_cols)

    # At this point, there should be no missing values left in the dataframe, unless
    # the DROP_ROW preprocessing option was selected, in which case we need to drop those
    # rows.
    len_dataset_before_drop_rows = len(dataset)
    dataset = dataset.dropna()
    len_dataset_after_drop_rows = len(dataset)

    if len_dataset_before_drop_rows != len_dataset_after_drop_rows:
        logger.warning(
            f"Dropped a total of {len_dataset_before_drop_rows - len_dataset_after_drop_rows} rows out of "
            f"{len_dataset_before_drop_rows} due to missing values"
        )

    # NaNs introduced by outer join change dtype of dataset cols (upcast to float64), so we need to cast them back.
    col_name_to_dtype = {}
    for col_name, col in proc_cols.items():
        # if col is a list of list-like objects, we assume the internal dtype of each col[i] remains unchanged.
        if type(col) == list and type(col[0]) in {list, np.ndarray, torch.Tensor}:
            continue
        col_name_to_dtype[col_name] = col.dtype
    dataset = dataset.astype(col_name_to_dtype)

    # Persist the completed dataset with no NaNs
    dataset = backend.df_engine.persist(dataset)

    # Remove partitions that are empty after removing NaNs
    dataset = backend.df_engine.remove_empty_partitions(dataset)

    # Embed features with fixed encoders
    dataset = embed_fixed_features(dataset, feature_configs, metadata, backend)

    return dataset, metadata


def embed_fixed_features(
    dataset: DataFrame, feature_configs: List[FeatureConfigDict], metadata: TrainingSetMetadataDict, backend: Backend
) -> DataFrame:
    """Transforms every input feature with cacheable encoder embeddings into its encoded form and updates
    metadata."""
    # Encode features in bulk at the end
    features_to_encode = get_features_with_cacheable_fixed_embeddings(feature_configs, metadata)
    if not features_to_encode:
        return dataset

    for feature in features_to_encode:
        # Temporarily set to False to ensure proper encoding
        metadata[feature[NAME]][PREPROCESSING]["cache_encoder_embeddings"] = False

    batch_size = backend.tune_batch_size(create_embed_batch_size_evaluator(features_to_encode, metadata), len(dataset))
    transform_fn = create_embed_transform_fn(features_to_encode, metadata)
    results = backend.batch_transform(dataset, batch_size, transform_fn)

    for feature in features_to_encode:
        # Set metadata so we know to skip encoding the feature
        metadata[feature[NAME]][PREPROCESSING]["cache_encoder_embeddings"] = True

    return results


def get_features_with_cacheable_fixed_embeddings(
    feature_configs: List[FeatureConfigDict], metadata: TrainingSetMetadataDict
) -> List[FeatureConfigDict]:
    """Returns list of features with `cache_encoder_embeddings=True` set in the preprocessing config."""
    features_to_encode = []
    for feature_config in feature_configs:
        # deal with encoders that have fixed preprocessing
        if ENCODER in feature_config:
            encoder_params = feature_config[ENCODER]
            if TYPE in encoder_params:
                preprocessing = metadata[feature_config[NAME]][PREPROCESSING]
                if preprocessing.get("cache_encoder_embeddings"):
                    # TODO(travis): passing in MODEL_ECD is a hack here that can be removed once we move to using
                    # the config object everywhere in preprocessing. Then we won't need to do the lookup on the
                    # encoder schema at all. This hack works for now because all encoders are supported by ECD, so
                    # there is no chance of a GBM model using an encoder not supported by ECD, but this could change
                    # in the future.
                    encoder_class = get_encoder_cls(MODEL_ECD, feature_config[TYPE], encoder_params[TYPE])
                    encoder = encoder_class.from_dict(encoder_params)
                    if not encoder.can_cache_embeddings():
                        raise ValueError(
                            f"Set `cache_encoder_embeddings=True` for feature {feature_config[NAME]} with "
                            f"encoder {encoder_params[TYPE]}, but encoder embeddings are not static."
                        )

                    # Convert to Ray Datasets, map batches to encode, then convert back to Dask
                    features_to_encode.append(feature_config)

    return features_to_encode


def cast_columns(dataset_cols, features, backend) -> None:
    """Casts columns based on their feature type."""
    for feature in features:
        # todo figure out if additional parameters are needed
        #  for the cast_column function
        try:
            dataset_cols[feature[COLUMN]] = get_from_registry(feature[TYPE], get_base_type_registry()).cast_column(
                dataset_cols[feature[COLUMN]], backend
            )
        except KeyError as e:
            raise KeyError(
                f"Feature name {e} specified in the config was not found in dataset with columns: "
                + f"{list(dataset_cols.keys())}"
            )


def merge_preprocessing(
    feature_config: FeatureConfigDict, global_preprocessing_parameters: PreprocessingConfigDict
) -> FeatureConfigDict:
    if PREPROCESSING not in feature_config:
        return global_preprocessing_parameters[feature_config[TYPE]]

    return merge_dict(global_preprocessing_parameters[feature_config[TYPE]], feature_config[PREPROCESSING])


def build_preprocessing_parameters(
    dataset_cols: Dict[str, Series],
    feature_configs: List[FeatureConfigDict],
    global_preprocessing_parameters: PreprocessingConfigDict,
    backend: Backend,
    metadata: Optional[TrainingSetMetadataDict] = None,
) -> PreprocessingConfigDict:
    if metadata is None:
        metadata = {}

    feature_name_to_preprocessing_parameters = {}
    for feature_config in feature_configs:
        feature_name = feature_config[NAME]

        # if metadata already exists, we can use it to get preprocessing parameters
        if feature_name in metadata:
            feature_name_to_preprocessing_parameters[feature_name] = metadata[feature_name][PREPROCESSING]
            continue

        preprocessing_parameters = feature_config[PREPROCESSING]
        missing_value_strategy = preprocessing_parameters["missing_value_strategy"]
        fill_value = precompute_fill_value(
            dataset_cols, feature_config, missing_value_strategy, preprocessing_parameters, backend
        )
        if fill_value is not None:
            preprocessing_parameters.update({"computed_fill_value": fill_value})

        # Handle outlier replacement
        outlier_strategy = preprocessing_parameters.get("outlier_strategy")
        if outlier_strategy is not None:
            if outlier_strategy != missing_value_strategy:
                outlier_fill_value = precompute_fill_value(
                    dataset_cols, feature_config, outlier_strategy, preprocessing_parameters, backend
                )
            else:
                # Use fill value from missing_value_strategy to avoid redundant computation
                outlier_fill_value = fill_value

            if outlier_fill_value is not None:
                preprocessing_parameters.update({"computed_outlier_fill_value": outlier_fill_value})

        feature_name_to_preprocessing_parameters[feature_name] = preprocessing_parameters

    return feature_name_to_preprocessing_parameters


def is_input_feature(feature_config: FeatureConfigDict) -> bool:
    """Utility function to check for the presence of encoder in the feature config to determine if the feature is
    an input feature or output feature."""
    return ENCODER in feature_config


def build_metadata(
    metadata: TrainingSetMetadataDict,
    feature_name_to_preprocessing_parameters: Dict[str, PreprocessingConfigDict],
    dataset_cols: Dict[str, Series],
    feature_configs: List[FeatureConfigDict],
    backend: Backend,
) -> TrainingSetMetadataDict:
    for feature_config in feature_configs:
        feature_name = feature_config[NAME]
        if feature_name in metadata:
            continue

        preprocessing_parameters = feature_name_to_preprocessing_parameters[feature_name]

        column = dataset_cols[feature_config[COLUMN]]
        metadata[feature_name] = get_from_registry(feature_config[TYPE], get_base_type_registry()).get_feature_meta(
            column, preprocessing_parameters, backend, is_input_feature(feature_config)
        )

        metadata[feature_name][PREPROCESSING] = preprocessing_parameters

    return metadata


def build_data(
    input_cols: DataFrame,
    feature_configs: List[Dict],
    training_set_metadata: Dict,
    backend: Backend,
    skip_save_processed_input: bool,
) -> Dict[str, DataFrame]:
    """Preprocesses the input dataframe columns, handles missing values, and potentially adds metadata to
    training_set_metadata.

    Args:
        input_cols: Input dataframe to be processed.
        feature_configs: List of feature configs.
        training_set_metadata: Training set metadata. Additional fields may be added.
        backend: Backend for data processing.
        skip_save_processed_input: (bool) Whether to skip saving the processed input.

    Returns:
        Dictionary of (feature name) -> (processed data).
    """
    proc_cols = {}
    for feature_config in feature_configs:
        preprocessing_parameters = training_set_metadata[feature_config[NAME]][PREPROCESSING]

        # Need to run this again here as cast_columns may have introduced new missing values
        handle_missing_values(input_cols, feature_config, preprocessing_parameters, backend)

        # For features that support it, we perform outlier removal here using metadata computed on the full dataset
        handle_outliers(
            input_cols, feature_config, preprocessing_parameters, training_set_metadata[feature_config[NAME]], backend
        )

        get_from_registry(feature_config[TYPE], get_base_type_registry()).add_feature_data(
            feature_config,
            input_cols,
            proc_cols,
            training_set_metadata,
            preprocessing_parameters,
            backend,
            skip_save_processed_input,
        )

    return proc_cols


def balance_data(
    dataset_df: DataFrame,
    output_features: List[Dict],
    preprocessing_parameters: Dict,
    backend: Backend,
    random_seed: int,
):
    """The purpose of this function is to balance the training dataset using either over-sampling or under-
    sampling.

    Args:
        dataset_df: Input dataframe to be over-sampled or under-sampled.
        output_features: List of feature configs.
        preprocessing_parameters: Dictionary of the global preprocessing parameters.
        backend: Backend for data processing.
        random_seed: Integer to seed the random sampling to ensure determinism.

    Returns: An over-sampled or under-sampled training dataset.
    """
    target = output_features[0][PROC_COLUMN]

    if backend.df_engine.partitioned:
        majority_class = backend.df_engine.compute(dataset_df[target].value_counts()).idxmax()
        minority_class = backend.df_engine.compute(dataset_df[target].value_counts()).idxmin()
    else:
        majority_class = dataset_df[target].value_counts().idxmax()
        minority_class = dataset_df[target].value_counts().idxmin()
    majority_df = dataset_df[dataset_df[target] == majority_class]
    minority_df = dataset_df[dataset_df[target] == minority_class]

    if preprocessing_parameters["oversample_minority"]:
        sample_fraction = (len(majority_df) * preprocessing_parameters["oversample_minority"]) / len(minority_df)
        minority_df = minority_df.sample(frac=sample_fraction, replace=True, random_state=random_seed)
    elif preprocessing_parameters["undersample_majority"]:
        sample_fraction = int(len(minority_df) / preprocessing_parameters["undersample_majority"]) / len(majority_df)
        majority_df = majority_df.sample(frac=sample_fraction, replace=False, random_state=random_seed)

    balanced_df = backend.df_engine.concat([minority_df, majority_df])

    return balanced_df


def precompute_fill_value(
    dataset_cols, feature, missing_value_strategy: str, preprocessing_parameters: PreprocessingConfigDict, backend
):
    """Precomputes the fill value for a feature.

    NOTE: this is called before NaNs are removed from the dataset. Modifications here must handle NaNs gracefully.
    NOTE: this is called before columns are cast. Modifications here must handle dtype conversion gracefully.
    """
    if missing_value_strategy == FILL_WITH_CONST:
        return preprocessing_parameters["fill_value"]
    elif missing_value_strategy == FILL_WITH_MODE:
        # Requires separate handling if Dask since Dask has lazy evaluation
        # Otherwise, dask returns a Dask index structure instead of a value to use as a fill value
        return (
            dataset_cols[feature[COLUMN]].value_counts().index.compute()[0]
            if is_dask_series_or_df(dataset_cols[feature[COLUMN]], backend)
            else dataset_cols[feature[COLUMN]].value_counts().index[0]
        )
    elif missing_value_strategy == FILL_WITH_MEAN:
        if feature[TYPE] != NUMBER:
            raise ValueError(
                f"Filling missing values with mean is supported "
                f"only for number types, not for type {feature[TYPE]}.",
            )
        return backend.df_engine.compute(dataset_cols[feature[COLUMN]].astype(float).mean())
    elif missing_value_strategy == FILL_WITH_FALSE:
        distinct_values = backend.df_engine.compute(
            dataset_cols[feature[COLUMN]].drop_duplicates().dropna()
        ).values.tolist()
        if len(distinct_values) > 2:
            raise ValueError(
                f"Missing value strategy `fill_with_false` "
                f"for column {feature[COLUMN]} expects 2 distinct values, "
                f"found: {len(distinct_values)} (ex: {distinct_values[:10]})"
            )

        # Determine the False label.
        # Distinct values are sorted in reverse to mirror the selection of the default fallback_true_label (in
        # binary_feature.get_feature_meta) for binary columns with unconventional boolean values, "human"/"bot".
        for v in sorted(distinct_values, reverse=True):
            fallback_true_label = (
                preprocessing_parameters["fallback_true_label"]
                # By default, preprocessing_parameters.fallback_true_label is None.
                if preprocessing_parameters["fallback_true_label"]
                else "true"
            )
            if strings_utils.str2bool(v, fallback_true_label) is False:
                return v
        raise ValueError(
            f"Unable to determine False value for column {feature[COLUMN]} with distinct values: {distinct_values}."
        )
    # Otherwise, we cannot precompute the fill value for this dataset
    return None


@DeveloperAPI
def handle_missing_values(dataset_cols, feature, preprocessing_parameters: PreprocessingConfigDict, backend):
    missing_value_strategy = preprocessing_parameters["missing_value_strategy"]
    computed_fill_value = preprocessing_parameters.get("computed_fill_value")
    _handle_missing_values(dataset_cols, feature, missing_value_strategy, computed_fill_value, backend)


@DeveloperAPI
def handle_outliers(dataset_cols, feature, preprocessing_parameters: PreprocessingConfigDict, metadata, backend):
    outlier_strategy = preprocessing_parameters.get("outlier_strategy")
    if outlier_strategy is None:
        return

    outlier_threshold = preprocessing_parameters["outlier_threshold"]
    computed_fill_value = preprocessing_parameters.get("computed_outlier_fill_value")

    # Identify all outliers and set them to NA so they can be removed
    series = dataset_cols[feature[COLUMN]]
    dataset_cols[feature[COLUMN]] = series.mask(
        series.sub(metadata["mean"]).div(metadata["std"]).abs().gt(outlier_threshold)
    )

    _handle_missing_values(dataset_cols, feature, outlier_strategy, computed_fill_value, backend)


def _handle_missing_values(
    dataset_cols, feature, missing_value_strategy: str, computed_fill_value: Optional[float], backend
):
    if (
        missing_value_strategy in {FILL_WITH_CONST, FILL_WITH_MODE, FILL_WITH_MEAN, FILL_WITH_FALSE}
        and computed_fill_value is not None
    ):
        dataset_cols[feature[COLUMN]] = dataset_cols[feature[COLUMN]].fillna(
            computed_fill_value,
        )
    elif missing_value_strategy in {BFILL, FFILL}:
        dataset_cols[feature[COLUMN]] = dataset_cols[feature[COLUMN]].fillna(
            method=missing_value_strategy,
        )

        # If the first few rows or last few rows of a dataset is a NaN, it will still be a NaN after ffill or bfill are
        # applied. This causes downstream errors with Dask (https://github.com/ludwig-ai/ludwig/issues/2452)
        # To get around this issue, apply the primary missing value strategy (say bfill) first, and then follow it
        # up with the other missing value strategy (ffill) to ensure all NaNs are filled
        if backend.df_engine.compute(dataset_cols[feature[COLUMN]].isna().sum()) > 0:
            dataset_cols[feature[COLUMN]] = dataset_cols[feature[COLUMN]].fillna(
                method=BFILL if missing_value_strategy == FFILL else FFILL,
            )
    elif missing_value_strategy == DROP_ROW:
        # Here we only drop from this series, but after preprocessing we'll do a second
        # round of dropping NA values from the entire output dataframe, which will
        # result in the removal of the rows.
        len_before_dropped_rows = len(dataset_cols[feature[COLUMN]])
        dataset_cols[feature[COLUMN]] = dataset_cols[feature[COLUMN]].dropna()
        len_after_dropped_rows = len(dataset_cols[feature[COLUMN]])

        if len_before_dropped_rows != len_after_dropped_rows:
            logger.warning(
                f"DROP_ROW missing value strategy applied. Dropped {len_before_dropped_rows - len_after_dropped_rows} "
                f"samples out of {len_before_dropped_rows} from column {feature[COLUMN]}. The rows containing these "
                f"samples will ultimately be dropped from the dataset."
            )
    else:
        raise ValueError(f"Invalid missing value strategy {missing_value_strategy}")


def load_hdf5(hdf5_file_path, preprocessing_params, backend, split_data=True, shuffle_training=False):
    # TODO dask: this needs to work with DataFrames
    logger.info(f"Loading data from: {hdf5_file_path}")

    def shuffle(df):
        return df.sample(frac=1).reset_index(drop=True)

    dataset = data_utils.load_hdf5(hdf5_file_path)
    if not split_data:
        if shuffle_training:
            dataset = shuffle(dataset)
        return dataset

    training_set, validation_set, test_set = split_dataset(dataset, preprocessing_params, backend)

    if shuffle_training:
        training_set = shuffle(training_set)

    return training_set, test_set, validation_set


def load_metadata(metadata_file_path: str) -> TrainingSetMetadataDict:
    logger.info(f"Loading metadata from: {metadata_file_path}")
    training_set_metadata = data_utils.load_json(metadata_file_path)
    # TODO(travis): decouple config from training_set_metadata so we don't need to
    #  upgrade it over time.
    training_set_metadata = upgrade_metadata(training_set_metadata)
    return training_set_metadata


def preprocess_for_training(
    config,
    dataset=None,
    training_set=None,
    validation_set=None,
    test_set=None,
    training_set_metadata=None,
    data_format=None,
    skip_save_processed_input=False,
    preprocessing_params=default_training_preprocessing_parameters,
    backend=LOCAL_BACKEND,
    random_seed=default_random_seed,
    callbacks=None,
) -> Tuple[Dataset, Dataset, Dataset, TrainingSetMetadataDict]:
    """Returns training, val and test datasets with training set metadata."""

    # sanity check to make sure some data source is provided
    if dataset is None and training_set is None:
        raise ValueError("No training data is provided!")

    # preload ludwig datasets
    dataset, training_set, validation_set, test_set = load_dataset_uris(
        dataset, training_set, validation_set, test_set, backend
    )

    # determine data format if not provided or auto
    if not data_format or data_format == "auto":
        data_format = figure_data_format(dataset, training_set, validation_set, test_set)

    # Wrap dataset into a form we can use to manage within the cache
    dataset = wrap(dataset)
    training_set = wrap(training_set)
    validation_set = wrap(validation_set)
    test_set = wrap(test_set)

    try:
        lock_path = backend.cache.get_cache_directory(dataset)
    except (TypeError, ValueError):
        lock_path = None
    with file_lock(lock_path, lock_file=".lock_preprocessing"):
        # if training_set_metadata is a string, assume it's a path to load the json
        training_set_metadata = training_set_metadata or {}
        if training_set_metadata and isinstance(training_set_metadata, str):
            training_set_metadata = load_metadata(training_set_metadata)

        # setup
        features = config["input_features"] + config["output_features"]

        # in case data_format is one of the cacheable formats,
        # check if there's a cached hdf5 file with the same name,
        # and in case move on with the hdf5 branch.
        cached = False
        cache = backend.cache.get_dataset_cache(config, dataset, training_set, test_set, validation_set)

        # Unwrap dataset into the form used for preprocessing
        dataset = dataset.unwrap() if dataset is not None else None
        training_set = training_set.unwrap() if training_set is not None else None
        validation_set = validation_set.unwrap() if validation_set is not None else None
        test_set = test_set.unwrap() if test_set is not None else None

        if data_format in CACHEABLE_FORMATS:
            with backend.storage.cache.use_credentials():
                # cache.get() returns valid indicating if the checksum for the current config
                # is equal to that from the cached training set metadata, as well as the paths to the
                # cached training set metadata, training set, validation_set, test set
                cache_results = cache.get()
                if cache_results is not None:
                    valid, *cache_values = cache_results
                    if valid:
                        logger.info(_get_cache_hit_message(cache))
                        training_set_metadata, training_set, test_set, validation_set = cache_values
                        config["data_hdf5_fp"] = training_set
                        data_format = backend.cache.data_format
                        cached = True
                        dataset = None
                    else:
                        logger.info(
                            "Found cached dataset and meta.json with the same filename "
                            "of the dataset, but checksums don't match, "
                            "if saving of processed input is not skipped "
                            "they will be overridden"
                        )
                        cache.delete()

        training_set_metadata[CHECKSUM] = cache.checksum
        data_format_processor = get_from_registry(data_format, data_format_preprocessor_registry)

        if cached or data_format == "hdf5":
            with backend.storage.cache.use_credentials():
                # Always interpret hdf5 files as preprocessed, even if missing from the cache
                processed = data_format_processor.prepare_processed_data(
                    features,
                    dataset=dataset,
                    training_set=training_set,
                    validation_set=validation_set,
                    test_set=test_set,
                    training_set_metadata=training_set_metadata,
                    skip_save_processed_input=skip_save_processed_input,
                    preprocessing_params=preprocessing_params,
                    backend=backend,
                    random_seed=random_seed,
                )
                training_set, test_set, validation_set, training_set_metadata = processed
        else:
            processed = data_format_processor.preprocess_for_training(
                config,
                features,
                dataset=dataset,
                training_set=training_set,
                validation_set=validation_set,
                test_set=test_set,
                training_set_metadata=training_set_metadata,
                skip_save_processed_input=skip_save_processed_input,
                preprocessing_params=preprocessing_params,
                backend=backend,
                random_seed=random_seed,
                callbacks=callbacks,
            )
            training_set, test_set, validation_set, training_set_metadata = processed
            processed = (training_set, test_set, validation_set, training_set_metadata)

            # cache the dataset
            if backend.cache.can_cache(skip_save_processed_input):
                with backend.storage.cache.use_credentials():
                    logger.debug("cache processed data")
                    processed = cache.put(*processed)
                    # set cached=True to ensure credentials are used correctly below
                    cached = True
            training_set, test_set, validation_set, training_set_metadata = processed

        with backend.storage.cache.use_credentials() if cached else contextlib.nullcontext():
            logger.debug("create training dataset")
            training_dataset = backend.dataset_manager.create(training_set, config, training_set_metadata)
            training_set_size = len(training_dataset)
            if training_set_size == 0:
                raise ValueError("Training data is empty following preprocessing.")
            elif training_set_size < MIN_DATASET_SPLIT_ROWS:
                raise ValueError(
                    f"Training dataset has only {training_set_size} rows following preprocessing, need"
                    f" at least {MIN_DATASET_SPLIT_ROWS} to compute metrics."
                )

            validation_dataset = None
            if validation_set is not None:
                logger.debug("create validation dataset")
                validation_dataset = backend.dataset_manager.create(validation_set, config, training_set_metadata)
                validation_set_size = len(validation_dataset)
                if validation_set_size == 0:
                    logger.warning(
                        "Validation set empty. If this is unintentional, please check the preprocessing configuration."
                    )
                    validation_dataset = None
                elif validation_set_size < MIN_DATASET_SPLIT_ROWS:
                    logger.warning(
                        f"Validation set too small to compute metrics. Need at least {MIN_DATASET_SPLIT_ROWS} rows, got"
                        f" {validation_set_size} after preprocessing."
                    )

            test_dataset = None
            if test_set is not None:
                logger.debug("create test dataset")
                test_dataset = backend.dataset_manager.create(test_set, config, training_set_metadata)
                test_set_size = len(test_dataset)
                if test_set_size == 0:
                    logger.warning(
                        "Test set empty. If this is unintentional, please check the preprocessing configuration."
                    )
                    test_dataset = None
                elif test_set_size < MIN_DATASET_SPLIT_ROWS:
                    logger.warning(
                        f"Test set too small to compute metrics. Need at least {MIN_DATASET_SPLIT_ROWS} rows, got"
                        f" {test_set_size} after preprocessing."
                    )

        return (training_dataset, validation_dataset, test_dataset, training_set_metadata)


def _preprocess_file_for_training(
    config,
    features,
    dataset=None,
    training_set=None,
    validation_set=None,
    test_set=None,
    training_set_metadata=None,
    read_fn=read_csv,
    skip_save_processed_input=False,
    preprocessing_params=default_training_preprocessing_parameters,
    backend=LOCAL_BACKEND,
    random_seed=default_random_seed,
    callbacks=None,
):
    """Method to pre-process csv data.

    :param features: list of all features (input + output)
    :param dataset: path to the data
    :param training_set:  training data
    :param validation_set: validation data
    :param test_set: test data
    :param training_set_metadata: train set metadata
    :param skip_save_processed_input: if False, the pre-processed data is saved
    as .hdf5 files in the same location as the csv files with the same names.
    :param preprocessing_params: preprocessing parameters
    :param random_seed: random seed
    :return: training, test, validation datasets, training metadata
    """
    if dataset:
        # Use data and ignore _train, _validation and _test.
        # Also ignore data and train set metadata needs preprocessing
        logger.info("Using full raw dataset, no hdf5 and json file " "with the same name have been found")
        logger.info("Building dataset (it may take a while)")

        dataset_df = read_fn(dataset, backend.df_engine.df_lib)
        training_set_metadata[SRC] = dataset

        data, training_set_metadata = build_dataset(
            dataset_df,
            features,
            preprocessing_params,
            mode="training",
            metadata=training_set_metadata,
            backend=backend,
            random_seed=random_seed,
            skip_save_processed_input=skip_save_processed_input,
            callbacks=callbacks,
        )

    elif training_set:
        # use data_train (including _validation and _test if they are present)
        # and ignore data and train set metadata
        # needs preprocessing
        logger.info("Using training raw csv, no hdf5 and json " "file with the same name have been found")
        logger.info("Building dataset (it may take a while)")

        concatenated_df = concatenate_files(training_set, validation_set, test_set, read_fn, backend)
        training_set_metadata[SRC] = training_set

        # Data is pre-split.
        preprocessing_params = set_fixed_split(preprocessing_params)

        data, training_set_metadata = build_dataset(
            concatenated_df,
            features,
            preprocessing_params,
            mode="training",
            metadata=training_set_metadata,
            backend=backend,
            random_seed=random_seed,
            callbacks=callbacks,
        )

    else:
        raise ValueError("either data or data_train have to be not None")

    logger.debug("split train-val-test")
    training_data, validation_data, test_data = split_dataset(data, preprocessing_params, backend, random_seed)

    if dataset and backend.is_coordinator() and not skip_save_processed_input:
        logger.debug("writing split file")
        splits_df = concatenate_splits(training_data, validation_data, test_data, backend)
        split_fp = get_split_path(dataset or training_set)
        try:
            backend.df_engine.to_parquet(splits_df, split_fp, index=True)
        except Exception as e:
            logger.warning(
                f"Encountered error: '{e}' while writing data to parquet during saving preprocessed data. "
                "Skipping saving processed data."
            )

    logger.info("Building dataset: DONE")
    if preprocessing_params["oversample_minority"] or preprocessing_params["undersample_majority"]:
        training_data = balance_data(
            training_data, config["output_features"], preprocessing_params, backend, random_seed
        )

    return training_data, test_data, validation_data, training_set_metadata


def _preprocess_df_for_training(
    config,
    features,
    dataset=None,
    training_set=None,
    validation_set=None,
    test_set=None,
    training_set_metadata=None,
    preprocessing_params=default_training_preprocessing_parameters,
    backend=LOCAL_BACKEND,
    random_seed=default_random_seed,
    callbacks=None,
):
    """Method to pre-process dataframes.

    This doesn't have the option to save the processed data as hdf5 as we don't expect users to do this as the data can
    be processed in memory
    """
    if dataset is not None:
        # needs preprocessing
        logger.info("Using full dataframe")
    elif training_set is not None:
        # needs preprocessing
        logger.info("Using training dataframe")
        dataset = concatenate_df(training_set, validation_set, test_set, backend)

        # Data is pre-split.
        preprocessing_params = set_fixed_split(preprocessing_params)

    logger.info("Building dataset (it may take a while)")

    data, training_set_metadata = build_dataset(
        dataset,
        features,
        preprocessing_params,
        mode="training",
        metadata=training_set_metadata,
        random_seed=random_seed,
        backend=backend,
        callbacks=callbacks,
    )

    logger.debug("split train-val-test")
    training_set, validation_set, test_set = split_dataset(data, preprocessing_params, backend, random_seed)

    logger.info("Building dataset: DONE")
    if preprocessing_params["oversample_minority"] or preprocessing_params["undersample_majority"]:
        training_set = balance_data(training_set, config["output_features"], preprocessing_params, backend, random_seed)

    return training_set, test_set, validation_set, training_set_metadata


def preprocess_for_prediction(
    config,
    dataset,
    training_set_metadata=None,
    data_format=None,
    split=FULL,
    include_outputs=True,
    backend=LOCAL_BACKEND,
    callbacks=None,
):
    """Preprocesses the dataset to parse it into a format that is usable by the Ludwig core.

    Args:
        config: Config dictionary corresponding to Ludwig Model
        dataset: Dataset to be processed
        training_set_metadata: Train set metadata for the input features
        data_format: Format of the data
        split: The split of dataset to return
        include_outputs: Whether to include outputs
        backend: Type of backend to use for preprocessing
        callbacks: Any callbacks passed in

    Returns:
        Processed dataset along with updated training set metadata
    """
    # Sanity Check to make sure some data source is provided
    if dataset is None:
        raise ValueError("No training data is provided!")

    if isinstance(dataset, Dataset):
        return dataset, training_set_metadata

    # preload ludwig datasets
    dataset, _, _, _ = load_dataset_uris(dataset, None, None, None, backend)

    # determine data format if not provided or auto
    if not data_format or data_format == "auto":
        data_format = figure_data_format(dataset)

    # manage the in_memory parameter
    if data_format not in HDF5_FORMATS:
        num_overrides = override_in_memory_flag(config["input_features"], True)
        if num_overrides > 0:
            logger.warning("Using in_memory = False is not supported " "with {} data format.".format(data_format))

    preprocessing_params = {}
    config_defaults = config.get(DEFAULTS, {})
    for feature_type in config_defaults:
        preprocessing_params[feature_type] = config_defaults[feature_type].get(PREPROCESSING, {})
    preprocessing_params[SPLIT] = config.get(PREPROCESSING, {}).get(SPLIT, {})

    preprocessing_params = merge_dict(default_prediction_preprocessing_parameters, preprocessing_params)

    # if training_set_metadata is a string, assume it's a path to load the json
    if training_set_metadata and isinstance(training_set_metadata, str):
        training_set_metadata = load_metadata(training_set_metadata)

    # setup
    output_features = []
    if include_outputs:
        output_features += config["output_features"]
    features = config["input_features"] + output_features

    # Check the cache for an already preprocessed dataset. This only
    # applies to scenarios where the user wishes to predict on a split
    # of the full dataset, where we preprocess the whole dataset together
    # during training. If the user wishes to predict on the full dataset,
    # it is assumed they are predicting on unseen data. This is done
    # because the cached data is stored in its split form, and would be
    # expensive to recombine, requiring further caching.
    cached = False

    dataset = wrap(dataset)
    cache = backend.cache.get_dataset_cache(config, dataset)
    dataset = dataset.unwrap()

    training_set = test_set = validation_set = None
    if data_format in CACHEABLE_FORMATS and split != FULL:
        with backend.storage.cache.use_credentials():
            cache_results = cache.get()
            if cache_results is not None:
                valid, *cache_values = cache_results
                if valid:
                    logger.info(_get_cache_hit_message(cache))
                    training_set_metadata, training_set, test_set, validation_set = cache_values
                    config["data_hdf5_fp"] = training_set
                    data_format = backend.cache.data_format
                    cached = True

    data_format_processor = get_from_registry(data_format, data_format_preprocessor_registry)
    if cached:
        with backend.storage.cache.use_credentials():
            processed = data_format_processor.prepare_processed_data(
                features,
                dataset=dataset,
                training_set=training_set,
                validation_set=validation_set,
                test_set=test_set,
                training_set_metadata=training_set_metadata,
                preprocessing_params=preprocessing_params,
                backend=backend,
            )
            training_set, test_set, validation_set, training_set_metadata = processed
    else:
        processed = data_format_processor.preprocess_for_prediction(
            dataset, features, preprocessing_params, training_set_metadata, backend, callbacks
        )
        dataset, training_set_metadata, new_hdf5_fp = processed
        training_set_metadata = training_set_metadata.copy()

        if new_hdf5_fp:
            training_set_metadata[DATA_TRAIN_HDF5_FP] = new_hdf5_fp

        if split != FULL:
            logger.debug("split train-val-test")
            training_set, validation_set, test_set = split_dataset(dataset, preprocessing_params, backend)

    if split == TRAINING:
        dataset = training_set
    elif split == VALIDATION:
        dataset = validation_set
    elif split == TEST:
        dataset = test_set

    config = {
        **config,
        "output_features": output_features,
    }

    with backend.storage.cache.use_credentials() if cached else contextlib.nullcontext():
        dataset = backend.dataset_manager.create(
            dataset,
            config,
            training_set_metadata,
        )

    return dataset, training_set_metadata


def _get_cache_hit_message(cache: DatasetCache) -> str:
    return (
        "Found cached dataset and meta.json with the same filename of the dataset.\n"
        "Using cached values instead of preprocessing the dataset again.\n"
        f"- Cached training set metadata path: {cache.get_cached_obj_path(META)}\n"
        f"- Cached training set path: {cache.get_cached_obj_path(TRAINING)}\n"
        f"- Cached validation set path: {cache.get_cached_obj_path(VALIDATION)}\n"
        f"- Cached test set path: {cache.get_cached_obj_path(TEST)}"
    )



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import H3, MODEL_ECD
from ludwig.schema.encoders.base import BaseEncoderConfig
from ludwig.schema.encoders.utils import EncoderDataclassField
from ludwig.schema.features.base import BaseInputFeatureConfig
from ludwig.schema.features.preprocessing.base import BasePreprocessingConfig
from ludwig.schema.features.preprocessing.utils import PreprocessingDataclassField
from ludwig.schema.features.utils import defaults_config_registry, ecd_input_config_registry, input_mixin_registry
from ludwig.schema.utils import BaseMarshmallowConfig, ludwig_dataclass


@DeveloperAPI
@defaults_config_registry.register(H3)
@input_mixin_registry.register(H3)
@ludwig_dataclass
class H3InputFeatureConfigMixin(BaseMarshmallowConfig):
    """H3InputFeatureConfigMixin is a dataclass that configures the parameters used in both the h3 input feature
    and the h3 global defaults section of the Ludwig Config."""

    preprocessing: BasePreprocessingConfig = PreprocessingDataclassField(feature_type=H3)

    encoder: BaseEncoderConfig = EncoderDataclassField(
        MODEL_ECD,
        feature_type=H3,
        default="embed",
    )


@DeveloperAPI
@ecd_input_config_registry.register(H3)
@ludwig_dataclass
class H3InputFeatureConfig(BaseInputFeatureConfig, H3InputFeatureConfigMixin):
    """H3InputFeatureConfig is a dataclass that configures the parameters used for an h3 input feature."""

    pass



import copy
import gc
import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import numpy as np
import numpy.typing as npt
import pandas as pd
import torch
import torch.nn as nn
from captum.attr import LayerIntegratedGradients, TokenReferenceBase
from captum.attr._utils.input_layer_wrapper import InputIdentity
from torch.autograd import Variable
from tqdm import tqdm

from ludwig.api import LudwigModel
from ludwig.api_annotations import PublicAPI
from ludwig.constants import CATEGORY, DATE, IMAGE, INPUT_FEATURES, NAME, PREPROCESSING, SET, TEXT, UNKNOWN_SYMBOL
from ludwig.data.preprocessing import preprocess_for_prediction
from ludwig.explain.explainer import Explainer
from ludwig.explain.explanation import ExplanationsResult
from ludwig.explain.util import get_pred_col
from ludwig.features.feature_utils import LudwigFeatureDict
from ludwig.models.ecd import ECD
from ludwig.utils.torch_utils import DEVICE

logger = logging.getLogger(__name__)


@dataclass
class ExplanationRunConfig:
    """Mutable state containing runtime configuration for explanation process.

    This is useful for updating the batch size used during explanation so it can be propagated across calls to
    `get_total_attribution`.
    """

    batch_size: int


def retry_on_cuda_oom(run_config: ExplanationRunConfig):
    def retry_on_cuda_oom_fn(fn):
        def retry_on_cuda_oom_wrapper(*args, **kwargs):
            while run_config.batch_size >= 2:
                try:
                    return fn(*args, **kwargs)
                except RuntimeError as e:
                    # PyTorch only generates Runtime errors for CUDA OOM.
                    gc.collect()
                    if "CUDA out of memory" in str(e) or isinstance(e, torch.cuda.OutOfMemoryError):
                        logger.exception(f"OOM at batch_size={run_config.batch_size}, halving and trying again")
                        run_config.batch_size //= 2
                    else:
                        # Not a CUDA error
                        raise

            raise RuntimeError("CUDA OOM raised during explanation, but batch size cannot be reduced any further")

        return retry_on_cuda_oom_wrapper

    return retry_on_cuda_oom_fn


class WrapperModule(torch.nn.Module):
    """Model used by the explainer to generate predictions.

    Unlike Ludwig's ECD class, this wrapper takes individual args as inputs to the forward function. We derive the order
    of these args from the order of the input_feature keys in ECD, which is guaranteed to be consistent (Python
    dictionaries are ordered consistently), so we can map back to the input feature dictionary as a second step within
    this wrapper.
    """

    def __init__(self, model: ECD, target: str):
        super().__init__()
        self.model = model
        self.target = target
        self.input_maps = nn.ModuleDict(
            {
                arg_name: InputIdentity(arg_name)
                for arg_name in self.model.input_features.keys()
                if self.model.input_features.get(arg_name).type() not in {TEXT, CATEGORY, DATE, SET}
            }
        )

    def forward(self, *args):
        # Add back the dictionary structure so it conforms to ECD format.
        input_features: LudwigFeatureDict = self.model.input_features
        inputs = {
            # Send the input through the identity layer so that we can use the output of the layer for attribution.
            # Except for text/category features where we use the embedding layer for attribution.
            feat_name: feat_input
            if input_features.get(feat_name).type() in {TEXT, CATEGORY, DATE, SET}
            else self.input_maps[feat_name](feat_input)
            for feat_name, feat_input in zip(input_features.keys(), args)
        }

        outputs = self.model(inputs)

        # At this point we only have the raw logits, but to make explainability work we need the probabilities
        # and predictions as well, so derive them.
        predictions = {}
        for of_name in self.model.output_features:
            predictions[of_name] = self.model.output_features.get(of_name).predictions(outputs, of_name)

        return get_pred_col(predictions, self.target)


@PublicAPI(stability="experimental")
class IntegratedGradientsExplainer(Explainer):
    def explain(self) -> ExplanationsResult:
        """Explain the model's predictions using Integrated Gradients.

        # Return

        :return: ExplanationsResult containing the explanations.
            `global_explanations`: (Explanation) Aggregate explanation for the entire input data.

            `row_explanations`: (List[Explanation]) A list of explanations, one for each row in the input data. Each
            explanation contains the integrated gradients for each label in the target feature's vocab with respect to
            each input feature.

            `expected_values`: (List[float]) of length [output feature cardinality] Average convergence delta for each
            label in the target feature's vocab.
        """

        # TODO(travis): add back skip encoders at the end in finally. Shouldn't be an issue in most cases as we
        # typically perform explanations on a loaded model and don't use it to predict afterwards.
        self.model.model.unskip()
        self.model.model.to(DEVICE)

        input_features: LudwigFeatureDict = self.model.model.input_features
        run_config = ExplanationRunConfig(batch_size=self.model.config_obj.trainer.batch_size)

        get_input_tensors_with_retry = retry_on_cuda_oom(run_config)(get_input_tensors)
        get_total_attribution_with_retry = retry_on_cuda_oom(run_config)(get_total_attribution)

        # Convert input data into embedding tensors from the output of the model encoders.
        inputs_encoded = get_input_tensors_with_retry(self.model, self.inputs_df, run_config)
        sample_encoded = get_input_tensors_with_retry(self.model, self.sample_df, run_config)
        baseline = get_baseline(self.model, sample_encoded)

        # Compute attribution for each possible output feature label separately.
        expected_values = []
        for target_idx in tqdm(range(self.vocab_size), desc="Explain"):
            total_attribution, feat_to_token_attributions, total_attribution_global = get_total_attribution_with_retry(
                self.model,
                self.target_feature_name,
                target_idx if self.is_category_target else None,
                inputs_encoded,
                baseline,
                len(self.inputs_df),
                run_config,
            )

            # Aggregate token attributions
            feat_to_token_attributions_global = {}
            for feat_name, token_attributions in feat_to_token_attributions.items():
                token_attributions_global = defaultdict(float)
                # sum attributions for each token
                for token, token_attribution in (ta for tas in token_attributions for ta in tas):
                    token_attributions_global[token] += abs(token_attribution)
                # divide by number of samples to get average attribution per token
                token_attributions_global = {
                    token: token_attribution / max(0, len(token_attributions))
                    for token, token_attribution in token_attributions_global.items()
                }
                # convert to list of tuples and sort by attribution
                token_attributions_global = sorted(token_attributions_global.items(), key=lambda x: x[1], reverse=True)
                # keep only top 100 tokens
                token_attributions_global = token_attributions_global[:100]
                feat_to_token_attributions_global[feat_name] = token_attributions_global

            self.global_explanation.add(
                input_features.keys(), total_attribution_global, feat_to_token_attributions_global
            )

            for i, (feature_attributions, explanation) in enumerate(zip(total_attribution, self.row_explanations)):
                # Add the feature attributions to the explanation object for this row.
                explanation.add(
                    input_features.keys(),
                    feature_attributions,
                    {k: v[i] for k, v in feat_to_token_attributions.items()},
                )

            # TODO(travis): for force plots, need something similar to SHAP E[X]
            expected_values.append(0.0)

            if self.is_binary_target:
                # For binary targets, we only need to compute attribution for the positive class (see below).
                break

        # For binary targets, add an extra attribution for the negative class (false).
        if self.is_binary_target:
            le_true = self.global_explanation.label_explanations[0]
            negated_attributions = le_true.to_array() * -1
            negated_token_attributions = {
                fa.feature_name: [(t, -a) for t, a in fa.token_attributions]
                for fa in le_true.feature_attributions
                if fa.token_attributions is not None
            }
            # Prepend the negative class to the list of label explanations.
            self.global_explanation.add(
                input_features.keys(), negated_attributions, negated_token_attributions, prepend=True
            )

            for explanation in self.row_explanations:
                le_true = explanation.label_explanations[0]
                negated_attributions = le_true.to_array() * -1
                negated_token_attributions = {
                    fa.feature_name: [(t, -a) for t, a in fa.token_attributions]
                    for fa in le_true.feature_attributions
                    if fa.token_attributions is not None
                }
                # Prepend the negative class to the list of label explanations.
                explanation.add(input_features.keys(), negated_attributions, negated_token_attributions, prepend=True)

            # TODO(travis): for force plots, need something similar to SHAP E[X]
            expected_values.append(0.0)

        return ExplanationsResult(self.global_explanation, self.row_explanations, expected_values)


def get_input_tensors(
    model: LudwigModel, input_set: pd.DataFrame, run_config: ExplanationRunConfig
) -> List[torch.Tensor]:
    """Convert the input data into a list of variables, one for each input feature.

    # Inputs

    :param model: The LudwigModel to use for encoding.
    :param input_set: The input data to encode of shape [batch size, num input features].

    # Return

    :return: A list of variables, one for each input feature. Shape of each variable is [batch size, embedding size].
    """
    # Ignore sample_ratio from the model config, since we want to explain all the data.
    sample_ratio_bak = model.config_obj.preprocessing.sample_ratio
    model.config_obj.preprocessing.sample_ratio = 1.0

    config = model.config_obj.to_dict()
    training_set_metadata = copy.deepcopy(model.training_set_metadata)
    for feature in config[INPUT_FEATURES]:
        preprocessing = training_set_metadata[feature[NAME]][PREPROCESSING]
        if preprocessing.get("cache_encoder_embeddings"):
            preprocessing["cache_encoder_embeddings"] = False

    # Convert raw input data into preprocessed tensor data
    dataset, _ = preprocess_for_prediction(
        config,
        dataset=input_set,
        training_set_metadata=training_set_metadata,
        data_format="auto",
        split="full",
        include_outputs=False,
        backend=model.backend,
        callbacks=model.callbacks,
    )

    # Restore sample_ratio
    model.config_obj.preprocessing.sample_ratio = sample_ratio_bak

    # Make sure the number of rows in the preprocessed dataset matches the number of rows in the input data
    assert (
        dataset.to_df().shape[0] == input_set.shape[0]
    ), f"Expected {input_set.shape[0]} rows in preprocessed dataset, but got {dataset.to_df().shape[0]}"

    # Convert dataset into a dict of tensors, and split each tensor into batches to control GPU memory usage
    inputs = {
        name: torch.from_numpy(dataset.dataset[feature.proc_column]).split(run_config.batch_size)
        for name, feature in model.model.input_features.items()
    }

    # Dict of lists to list of dicts
    input_batches = [dict(zip(inputs, t)) for t in zip(*inputs.values())]

    # List of dicts to dict of lists
    preproc_inputs = {k: torch.cat([d[k] for d in input_batches]) for k in input_batches[0]}

    data_to_predict = [v for _, v in preproc_inputs.items()]
    tensors = []
    for t in data_to_predict:
        # TODO(travis): Consider changing to `if not torch.is_floating_point(t.dtype)` to simplify, then handle bool
        # case in this block.
        if t.dtype == torch.int8 or t.dtype == torch.int16 or t.dtype == torch.int32 or t.dtype == torch.int64:
            # Don't wrap input into a variable if it's an integer type, since it will be used as an index into the
            # embedding table. We explain the output of the embedding table, not the input to the embedding table using
            # LayerIntegratedGradients.
            tensors.append(t)
        else:
            # Wrap input into a variable so torch will track the gradient and LayerIntegratedGradients can explain it.
            if t.dtype == torch.bool:
                t = t.to(torch.float32)
            tensors.append(Variable(t, requires_grad=True))

    return tensors


def get_baseline(model: LudwigModel, sample_encoded: List[Variable]) -> List[torch.Tensor]:
    # TODO(travis): pre-compute this during training from the full training dataset.
    input_features: LudwigFeatureDict = model.model.input_features

    baselines = []
    for sample_input, (name, feature) in zip(sample_encoded, input_features.items()):
        metadata = model.training_set_metadata[name]
        if feature.type() == TEXT:
            PAD_IND = metadata.get("pad_idx", metadata.get("word_pad_idx"))
            token_reference = TokenReferenceBase(reference_token_idx=PAD_IND)
            baseline = token_reference.generate_reference(sequence_length=sample_input.shape[1], device=DEVICE)
        elif feature.type() == CATEGORY:
            most_popular_token = max(metadata["str2freq"], key=metadata["str2freq"].get)
            most_popular_tok_idx = metadata["str2idx"].get(most_popular_token)

            # If an unknown is defined, use that as the baseline index, else use the most popular token
            baseline_tok_idx = metadata["str2idx"].get(UNKNOWN_SYMBOL, most_popular_tok_idx)
            baseline = torch.tensor(baseline_tok_idx, device=DEVICE)
        elif feature.type() == IMAGE:
            baseline = torch.zeros_like(sample_input[0], device=DEVICE)
        else:
            # For a robust baseline, we take the mean of all samples from the training data.
            baseline = torch.mean(sample_input.float(), dim=0)
        baselines.append(baseline.unsqueeze(0))

    return baselines


def get_total_attribution(
    model: LudwigModel,
    target_feature_name: str,
    target_idx: Optional[int],
    feature_inputs: List[Variable],
    baseline: List[torch.Tensor],
    nsamples: int,
    run_config: ExplanationRunConfig,
) -> Tuple[npt.NDArray[np.float64], Dict[str, List[List[Tuple[str, float]]]]]:
    """Compute the total attribution for each input feature for each row in the input data.

    Args:
        model: The Ludwig model to explain.
        target_feature_name: The name of the target feature to explain.
        target_idx: The index of the target feature label to explain if the target feature is a category.
        feature_inputs: The preprocessed input data as a list of tensors of length [num_features].
        baseline: The baseline input data as a list of tensors of length [num_features].
        nsamples: The total number of samples in the input data.

    Returns:
        The token-attribution pair for each token in the input feature for each row in the input data. The members of
        the output tuple are structured as follows:

        `total_attribution_rows`: (npt.NDArray[np.float64]) of shape [num_rows, num_features]
        The total attribution for each input feature for each row in the input data.

        `feat_to_token_attributions`: (Dict[str, List[List[Tuple[str, float]]]]) with values of shape
        [num_rows, seq_len, 2]

        `total_attribution_global`: (npt.NDArray[np.float64]) of shape [num_features]
        The attribution for each input feature aggregated across all input data.
    """
    input_features: LudwigFeatureDict = model.model.input_features

    # Configure the explainer, which includes wrapping the model so its interface conforms to
    # the format expected by Captum.
    model.model.zero_grad()
    explanation_model = WrapperModule(model.model, target_feature_name)

    layers = []
    for feat_name, feat in input_features.items():
        if feat.type() in {TEXT, CATEGORY, DATE, SET}:
            # Get embedding layer from encoder, which is the first child of the encoder.
            layers.append(feat.encoder_obj.get_embedding_layer())
        else:
            # Get the wrapped input layer.
            layers.append(explanation_model.input_maps[feat_name])

    explainer = LayerIntegratedGradients(explanation_model, layers)

    feature_inputs_splits = [ipt.split(run_config.batch_size) for ipt in feature_inputs]
    baseline = [t.to(DEVICE) for t in baseline]

    total_attribution_rows = None
    total_attribution_global = None
    feat_to_token_attributions = defaultdict(list)
    for input_batch in zip(*feature_inputs_splits):
        input_batch = [ipt.to(DEVICE) for ipt in input_batch]
        attribution = explainer.attribute(
            tuple(input_batch),
            baselines=tuple(baseline),
            target=target_idx,
            # https://captum.ai/docs/faq#i-am-facing-out-of-memory-oom-errors-when-using-captum-how-do-i-resolve-this
            internal_batch_size=run_config.batch_size,
        )

        attributions_reduced = []
        for a in attribution:
            a_reduced = a.detach().cpu()
            if a_reduced.ndim == 2 or a_reduced.ndim == 3:
                # Reduces category-level attributions of shape [batch_size, embedding_dim] by summing over the
                # embedding dimension to get attributions of shape [batch_size].
                # Reduces token-level attributions of shape [batch_size, sequence_length, embedding_dim] by summing
                # over the embedding dimension to get attributions of shape [batch_size, sequence_length]. We keep
                # the sequence dimension so we can map the attributions to the tokens.
                a_reduced = a_reduced.sum(dim=-1)
            elif a_reduced.ndim == 4:
                # Reduce pixel-level attributions of shape [batch_size, num_channels, height, width] by summing
                # over the channel and spatial dimensions to get attributions of shape [batch_size].
                a_reduced = a_reduced.sum(dim=(1, 2, 3))
            attributions_reduced.append(a_reduced)

        for inputs, attrs, (name, feat) in zip(input_batch, attributions_reduced, input_features.items()):
            if feat.type() == TEXT:
                tok_attrs = get_token_attributions(model, name, inputs.detach().cpu(), attrs)
                feat_to_token_attributions[name].append(tok_attrs)

        # Reduce attribution to [num_input_features, batch_size] by summing over the sequence dimension (if present).
        attribution = [a.sum(dim=-1) if a.ndim == 2 else a for a in attributions_reduced]
        attribution = np.stack(attribution)

        # Transpose to [batch_size, num_input_features]
        attribution = attribution.T

        if total_attribution_rows is not None:
            total_attribution_rows = np.concatenate([total_attribution_rows, attribution], axis=0)
        else:
            total_attribution_rows = attribution

        if total_attribution_global is not None:
            total_attribution_global += attribution.sum(axis=0)
        else:
            total_attribution_global = attribution.sum(axis=0)

    total_attribution_global /= nsamples

    feat_to_token_attributions = {k: [e for lst in v for e in lst] for k, v in feat_to_token_attributions.items()}

    return total_attribution_rows, feat_to_token_attributions, total_attribution_global


def get_token_attributions(
    model: LudwigModel,
    feature_name: str,
    input_ids: torch.Tensor,
    token_attributions: torch.Tensor,
) -> List[List[Tuple[str, float]]]:
    """Convert token-level attributions to an array of token-attribution pairs of shape.

    [batch_size, sequence_length, 2].

    Args:
        model: The LudwigModel used to generate the attributions.
        feature_name: The name of the feature for which the attributions were generated.
        input_ids: The input ids of shape [batch_size, sequence_length].
        token_attributions: The token-level attributions of shape [batch_size, sequence_length].

    Returns:
        An array of token-attribution pairs of shape [batch_size, sequence_length, 2].
    """
    assert (
        input_ids.dtype == torch.int8
        or input_ids.dtype == torch.int16
        or input_ids.dtype == torch.int32
        or input_ids.dtype == torch.int64
    )

    # Normalize token-level attributions to visualize the relative importance of each token.
    norm = torch.linalg.norm(token_attributions, dim=1)
    # Safe divide by zero by setting the norm to 1 if the norm is 0.
    norm = torch.where(norm == 0, torch.ones_like(norm), norm)
    token_attributions = token_attributions / norm.unsqueeze(-1)

    # map input ids to input tokens via the vocabulary
    feature = model.training_set_metadata[feature_name]
    vocab = feature.get("idx2str", feature.get("word_idx2str"))
    idx2str = np.vectorize(lambda idx: vocab[idx])
    input_tokens = idx2str(input_ids)

    # add attribution to the input tokens
    tok_attrs = [
        list(zip(t, a)) for t, a in zip(input_tokens, token_attributions.tolist())
    ]  # [batch_size, sequence_length, 2]

    return tok_attrs



import os
from contextlib import contextmanager
from typing import Optional

from ludwig.utils.fs_utils import upload_output_directory


def create_kaggle_client():
    # Need to import here to prevent Kaggle from authenticating on import
    from kaggle import api

    return api


@contextmanager
def update_env(**kwargs):
    override_env = {k: v for k, v in kwargs.items() if v is not None}
    old = os.environ.copy()
    try:
        os.environ.update(override_env)
        yield
    finally:
        os.environ = old


def download_kaggle_dataset(
    download_directory: str,
    kaggle_dataset_id: Optional[str] = None,
    kaggle_competition: Optional[str] = None,
    kaggle_username: Optional[str] = None,
    kaggle_key: Optional[str] = None,
):
    """Download all files in a kaggle dataset. One of kaggle_dataset_id,

    If the user has not specified creds in the kaggle.json file we lookup the passed in username and the api key and
    perform authentication.
    """
    with update_env(KAGGLE_USERNAME=kaggle_username, KAGGLE_KEY=kaggle_key):
        # Call authenticate explicitly to pick up new credentials if necessary
        api = create_kaggle_client()
        api.authenticate()
    with upload_output_directory(download_directory) as (tmpdir, _):
        if kaggle_competition:
            api.competition_download_files(kaggle_competition, path=tmpdir)
        else:
            api.dataset_download_files(kaggle_dataset_id, path=tmpdir)
    return [os.path.join(download_directory, f) for f in os.listdir(download_directory)]



from copy import deepcopy
from typing import Dict

import numpy as np
import pytest
import torch

from ludwig.features.number_feature import _OutlierReplacer, NumberInputFeature
from ludwig.schema.features.number_feature import ECDNumberInputFeatureConfig
from ludwig.schema.utils import load_config_with_kwargs
from ludwig.utils.misc_utils import merge_dict
from ludwig.utils.torch_utils import get_torch_device

BATCH_SIZE = 2
DEVICE = get_torch_device()


@pytest.fixture(scope="module")
def number_config():
    return {"name": "number_column_name", "type": "number"}


def test_number_input_feature(
    number_config: Dict,
) -> None:
    # setup image input feature definition
    number_def = deepcopy(number_config)

    # pickup any other missing parameters
    defaults = ECDNumberInputFeatureConfig().to_dict()
    set_def = merge_dict(defaults, number_def)

    # ensure no exceptions raised during build
    number_config, _ = load_config_with_kwargs(ECDNumberInputFeatureConfig, set_def)
    input_feature_obj = NumberInputFeature(number_config).to(DEVICE)

    # check one forward pass through input feature
    input_tensor = input_feature_obj.create_sample_input(batch_size=BATCH_SIZE)
    assert input_tensor.shape == torch.Size([BATCH_SIZE])
    assert input_tensor.dtype == torch.float32

    encoder_output = input_feature_obj(input_tensor)
    assert encoder_output["encoder_output"].shape == (BATCH_SIZE, *input_feature_obj.output_shape)


def test_outlier_replacer():
    replacer = _OutlierReplacer(
        {"mean": 50, "std": 30, "preprocessing": {"outlier_threshold": 2.0, "computed_outlier_fill_value": 42}}
    )

    t = torch.from_numpy(np.array([10, 20, 1000, -500, 80], dtype=np.float32))
    t_out_expected = torch.from_numpy(np.array([10, 20, 42, 42, 80], dtype=np.float32))

    t_out = replacer(t)
    assert torch.equal(t_out, t_out_expected)



import logging
from typing import Dict

import torch

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import HIDDEN, LOGITS, SEQUENCE, TEXT
from ludwig.decoders.base import Decoder
from ludwig.decoders.registry import register_decoder
from ludwig.modules.attention_modules import MultiHeadSelfAttention
from ludwig.schema.decoders.sequence_decoders import SequenceTaggerDecoderConfig
from ludwig.utils.torch_utils import Dense

logger = logging.getLogger(__name__)


@DeveloperAPI
@register_decoder("tagger", [SEQUENCE, TEXT])
class SequenceTaggerDecoder(Decoder):
    def __init__(
        self,
        input_size: int,
        vocab_size: int,
        max_sequence_length: int,
        use_attention: bool = False,
        use_bias: bool = True,
        attention_embedding_size: int = 256,
        attention_num_heads: int = 8,
        decoder_config=None,
        **kwargs,
    ):
        super().__init__()
        self.config = decoder_config

        self.vocab_size = vocab_size
        self.max_sequence_length = max_sequence_length
        self.input_size = input_size
        self.use_attention = use_attention
        if use_attention:
            logger.debug("  MultiHeadSelfAttention")
            self.self_attention = MultiHeadSelfAttention(
                input_size=input_size, hidden_size=attention_embedding_size, num_heads=attention_num_heads
            )
            # Adjust the input size to the final projection layer.
            input_size = self.self_attention.output_shape[0]
        self.projection_layer = Dense(input_size=input_size, output_size=vocab_size, use_bias=use_bias)

    def forward(self, inputs: Dict[str, torch.Tensor], target: torch.Tensor = None) -> Dict[str, torch.Tensor]:
        """Decodes the inputs into a sequence.

        Args:
            inputs: Dictionary of tensors from the outputs of the combiner and other output features.
            target: Tensor [batch_size, max_sequence_length] with predictions.

        Returns:
            Dictionary of tensors with logits [batch_size, max_sequence_length, vocab_size].
        """
        hidden = inputs[HIDDEN]
        if len(hidden.size()) != 3:
            raise ValueError(
                f"Decoder inputs rank is {len(hidden.size())}, but should be 3: "
                + "[batch_size x max_sequence_length x hidden_size] in when using a tagger sequential decoder. "
                + "Consider setting reduce_output to None if a sequential encoder / combiner is used."
            )
        if list(hidden.shape[1:]) != [self.max_sequence_length, self.input_size]:
            raise ValueError(
                "Sequence tagger decoder inputs (hidden) should be [batch_size, self.max_sequence_length, "
                + f"input_size], or [batch_size, {self.max_sequence_length}, {self.input_size}]. However, the "
                + f"inputs (hidden) was instead: {list(hidden.size())}. "
                + "The encoder is not length preserving. Please check its configuration."
            )

        if self.use_attention:
            hidden = self.self_attention(hidden)

        logits = self.projection_layer(hidden)
        return {LOGITS: logits}

    def get_prediction_set(self):
        return {LOGITS}

    @staticmethod
    def get_schema_cls():
        return SequenceTaggerDecoderConfig

    @property
    def input_shape(self):
        # Dummy implementation.
        return torch.Size([1])

    @property
    def output_shape(self):
        return torch.Size([self.max_sequence_length, self.vocab_size])



import logging
import os
import subprocess
import sys

import pytest

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.getLogger("ludwig").setLevel(logging.INFO)

TEST_SCRIPT = os.path.join(os.path.dirname(__file__), "scripts", "run_train_aim.py")


@pytest.mark.distributed
def test_contrib_experiment(csv_filename, tmpdir):
    aim_test_path = os.path.join(tmpdir, "results")
    os.makedirs(aim_test_path, exist_ok=True)

    os.environ["AIM_TEST_PATH"] = aim_test_path
    subprocess.call(["chmod", "-R", "a+w", os.environ["AIM_TEST_PATH"]])

    cmdline = [sys.executable, TEST_SCRIPT, "--csv-filename", csv_filename]
    print(cmdline)
    exit_code = subprocess.call(" ".join(cmdline), shell=True, env=os.environ.copy())
    assert exit_code == 0



#! /usr/bin/env python
# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
import logging
from collections import defaultdict
from functools import total_ordering
from typing import Callable, Dict, List, Optional

from packaging import version as pkg_version

logger = logging.getLogger(__name__)


@total_ordering
class VersionTransformation:
    """Wrapper class for transformations to config dicts."""

    def __init__(self, transform: Callable[[Dict], Dict], version: str, prefixes: List[str] = None):
        """Constructor.

        Args:
            transform: A function or other callable from Dict -> Dict which returns a modified version of the config.
                       The callable may update the config in-place and return it, or return a new dict.
            version: The Ludwig version, should be the first version which requires this transform.
            prefixes: A list of config prefixes this transform should apply to, i.e. ["hyperopt"].  If not specified,
                      transform will be called with the entire config dictionary.
        """
        self.transform = transform
        self.version = version
        self.pkg_version = pkg_version.parse(version)
        self.prefixes = prefixes if prefixes else []

    def transform_config(self, config: Dict):
        """Transforms the sepcified config, returns the transformed config."""
        prefixes = self.prefixes if self.prefixes else [""]
        for prefix in prefixes:
            if prefix and (prefix not in config or not config[prefix]):
                # If the prefix is non-empty (transformation applies to a specific section), but the section is either
                # absent or empty, then skip.
                continue
            config = self.transform_config_with_prefix(config, prefix)
        return config

    def transform_config_with_prefix(self, config: Dict, prefix: Optional[str] = None) -> Dict:
        """Applied this version transformation to a specified prefix of the config, returns the updated config. If
        prefix names a list, i.e. "input_features", applies the transformation to each list element (input
        feature).

        Args:
            config: A config dictionary.
            prefix: An optional keypath prefix i.e. "input_features". If no prefix specified, transformation is applied
                    to config itself.

        Returns The updated config.
        """
        if prefix:
            components = prefix.split(".", 1)
            key = components[0]
            rest_of_prefix = components[1] if len(components) > 1 else ""
            if key in config:
                subsection = config[key]
                if isinstance(subsection, list):
                    config[key] = [
                        self.transform_config_with_prefix(v, prefix=rest_of_prefix) if isinstance(v, dict) else v
                        for v in subsection
                    ]
                elif isinstance(subsection, dict):
                    config[key] = self.transform_config_with_prefix(subsection, prefix=rest_of_prefix)
            return config
        else:
            # Base case: no prefix specified, pass entire dictionary to transform function.
            transformed_config = self.transform(config)
            if transformed_config is None:
                logger.error("Error: version transformation returned None. Check for missing return statement.")
            return transformed_config

    @property
    def max_prefix_length(self):
        """Returns the length of the longest prefix."""
        return max(len(prefix.split(".")) for prefix in self.prefixes) if self.prefixes else 0

    @property
    def longest_prefix(self):
        """Returns the longest prefix, or empty string if no prefixes specified."""
        prefixes = self.prefixes
        if not prefixes:
            return ""
        max_index = max(range(len(prefixes)), key=lambda i: prefixes[i])
        return prefixes[max_index]

    def __lt__(self, other):
        """Defines sort order of version transformations. Sorted by:

        - version (ascending)
        - max_prefix_length (ascending) Process outer config transformations before inner.
        - longest_prefix (ascending) Order alphabetically by prefix if max_prefix_length equal.
        """
        return (self.pkg_version, self.max_prefix_length, self.longest_prefix) < (
            other.pkg_version,
            other.max_prefix_length,
            other.longest_prefix,
        )

    def __repr__(self):
        return f'VersionTransformation(<function>, version="{self.version}", prefixes={repr(self.prefixes)})'


class VersionTransformationRegistry:
    """A registry of transformations which update versioned config files."""

    def __init__(self):
        self._registry = defaultdict(list)  # Maps version number to list of transformations.

    def register(self, transformation: VersionTransformation):
        """Registers a version transformation."""
        self._registry[transformation.version].append(transformation)

    def get_transformations(self, from_version: str, to_version: str) -> List[VersionTransformation]:
        """Filters transformations to create an ordered list of the config transformations from one version to
        another. All transformations returned have version st. from_version < version <= to_version.

        Args:
            from_version: The ludwig version of the input config.
            to_version: The version to update the config to (usually the current LUDWIG_VERSION).

        Returns an ordered list of transformations to apply to the config to update it.
        """
        from_version = pkg_version.parse(from_version)

        # Ignore pre-release, development versions. Otherwise transformations for upcoming releases will not be applied.
        to_version = pkg_version.parse(to_version)
        to_version = pkg_version.parse(f"{to_version.major}.{to_version.minor}")

        def in_range(v, to_version, from_version):
            v = pkg_version.parse(v)
            return from_version <= v <= to_version

        versions = [v for v in self._registry.keys() if in_range(v, to_version, from_version)]

        transforms = sorted(t for v in versions for t in self._registry[v])
        return transforms

    def update_config(self, config: Dict, from_version: str, to_version: str) -> Dict:
        """Applies the transformations from an older version to a newer version.

        Args:
            config: The config, created by ludwig at from_version.
            from_version: The version of ludwig which wrote the older config.
            to_version: The version of ludwig to update to (usually the current LUDWIG_VERSION).

        Returns The updated config after applying update transformations and updating the "ludwig_version" key.
        """
        transformations = self.get_transformations(from_version, to_version)
        updated_config = copy.deepcopy(config)
        for t in transformations:
            updated_config = t.transform_config(updated_config)
        updated_config["ludwig_version"] = to_version
        return updated_config



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import glob
import hashlib
import logging
import os
import shutil
import urllib
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple, Union
from urllib.parse import urlparse

import pandas as pd
from tqdm import tqdm

from ludwig.api_annotations import DeveloperAPI, PublicAPI
from ludwig.constants import SPLIT
from ludwig.datasets.archives import extract_archive, is_archive, list_archive
from ludwig.datasets.dataset_config import DatasetConfig
from ludwig.datasets.kaggle import download_kaggle_dataset
from ludwig.datasets.utils import model_configs_for_dataset
from ludwig.utils.strings_utils import make_safe_filename

logger = logging.getLogger(__name__)


@DeveloperAPI
class TqdmUpTo(tqdm):
    """Provides progress bar for `urlretrieve`.

    Taken from: https://gist.github.com/leimao/37ff6e990b3226c2c9670a2cd1e4a6f5
    """

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


@DeveloperAPI
def get_default_cache_location() -> str:
    """Returns a path to the default LUDWIG_CACHE location, or $HOME/.ludwig_cache."""
    if "LUDWIG_CACHE" in os.environ and os.environ["LUDWIG_CACHE"]:
        return os.environ["LUDWIG_CACHE"]
    else:
        return str(Path.home().joinpath(".ludwig_cache"))


def _list_of_strings(list_or_string: Union[str, List[str]]) -> List[str]:
    """Helper function to accept single string or lists in config."""
    return [list_or_string] if isinstance(list_or_string, str) else list_or_string


def _glob_multiple(pathnames: List[str], root_dir: str = None, recursive: bool = True) -> Set[str]:
    """Recursive glob multiple patterns, returns set of matches.

    Note: glob's root_dir argument was added in python 3.10, not using it for compatibility.
    """
    if root_dir:
        pathnames = [os.path.join(root_dir, p) for p in pathnames]
    return set().union(*[glob.glob(p, recursive=recursive) for p in pathnames])


def _sha256_digest(file_path) -> str:
    """Returns the sha256 digest for the specified file."""
    hash = hashlib.sha256()
    buffer = bytearray(hash.block_size * 1024)  # Attempts to read in multiples of the hash block size (64KB).
    mv = memoryview(buffer)
    with open(file_path, "rb", buffering=0) as f:
        for bytes_read in iter(lambda: f.readinto(mv), 0):
            hash.update(mv[:bytes_read])
    return hash.hexdigest()


@PublicAPI
class DatasetState(int, Enum):
    """The state of the dataset."""

    NOT_LOADED = 0
    DOWNLOADED = 1
    EXTRACTED = 2
    TRANSFORMED = 3


@PublicAPI
class DatasetLoader:
    """Base class that defines the default pipeline for loading a ludwig dataset.

    Clients will typically call load(), which processes the dataset according to the config.

    A dataset is processed in 4 phases:
        1. Download       - The dataset files are downloaded to the cache.
        2. Verify         - Hashes of downloaded files are verified.
        3. Extract        - The dataset files are extracted from an archive (may be a no-op if data is not archived).
        4. Transform      - The dataset is transformed into a format usable for training and is ready to load.
            a. Transform Files      (Files -> Files)
            b. Load Dataframe       (Files -> DataFrame)
            c. Transform Dataframe  (DataFrame -> DataFrame)
            d. Save Processed       (DataFrame -> File)

    The download and extract phases are run for each URL based on the URL type and file extension. After extraction, the
    full set of downloaded and extracted files are collected and passed as a list to the transform stage.

    The transform phase offers customization points for datasets which require preprocessing before they are usable for
    training.
    """

    def __init__(self, config: DatasetConfig, cache_dir: Optional[str] = None):
        """Constructor."""
        self.config = config
        self.cache_dir = cache_dir if cache_dir else get_default_cache_location()

    @property
    def name(self):
        """The name of the dataset."""
        return self.config.name

    @property
    def version(self):
        """The version of the dataset."""
        return self.config.version

    @property
    def is_kaggle_dataset(self) -> bool:
        return self.config.kaggle_dataset_id or self.config.kaggle_competition

    @property
    def download_dir(self) -> str:
        """Directory where all dataset artifacts are saved."""
        return os.path.join(self.cache_dir, f"{self.name}_{self.version}")

    @property
    def raw_dataset_dir(self) -> str:
        """Save path for raw data downloaded from the web."""
        return os.path.join(self.download_dir, "raw")

    @property
    def processed_dataset_dir(self) -> str:
        """Save path for processed data."""
        return os.path.join(self.download_dir, "processed")

    @property
    def processed_dataset_filename(self) -> str:
        """Filename for processed data."""
        return f"{make_safe_filename(self.config.name)}.parquet"

    @property
    def processed_dataset_path(self) -> str:
        """Save path to processed dataset file."""
        return os.path.join(self.processed_dataset_dir, self.processed_dataset_filename)

    @property
    def processed_temp_dir(self) -> str:
        """Save path for processed temp data."""
        return os.path.join(self.download_dir, "_processed")

    @property
    def state(self) -> DatasetState:
        """Dataset state."""
        if os.path.exists(self.processed_dataset_path):
            return DatasetState.TRANSFORMED
        if all([os.path.exists(os.path.join(self.raw_dataset_dir, filename)) for filename in self.download_filenames]):
            archive_filenames = [f for f in self.download_filenames if is_archive(f)]
            if archive_filenames:
                # Check to see if archive has been extracted.
                extracted_files = [
                    f for a in archive_filenames for f in list_archive(os.path.join(self.raw_dataset_dir, a))
                ]
                if all(os.path.exists(os.path.join(self.raw_dataset_dir, ef)) for ef in extracted_files):
                    return DatasetState.EXTRACTED
                else:
                    return DatasetState.DOWNLOADED
            # If none of the dataset download files are archives, skip extraction phase.
            return DatasetState.EXTRACTED
        return DatasetState.NOT_LOADED

    @property
    def download_urls(self) -> List[str]:
        return _list_of_strings(self.config.download_urls)

    @property
    def download_filenames(self) -> List[str]:
        """Filenames for downloaded files inferred from download_urls."""
        if self.config.archive_filenames:
            return _list_of_strings(self.config.archive_filenames)
        return [os.path.basename(urlparse(url).path) for url in self.download_urls]

    def description(self) -> str:
        """Returns human-readable description of the dataset."""
        return f"{self.config.name} {self.config.version}\n{self.config.description}"

    @property
    def model_configs(self) -> Dict[str, Dict]:
        """Returns a dictionary of built-in model configs for this dataset."""
        return model_configs_for_dataset(self.config.name)

    @property
    def best_model_config(self) -> Optional[Dict]:
        """Returns the best built-in model config for this dataset, or None."""
        return self.model_configs.get("best")

    @property
    def default_model_config(self) -> Optional[Dict]:
        """Returns the default built-in model config for this dataset.

        This is a good first model which should train in under 10m on a current laptop without GPU acceleration.
        """
        return self.model_configs.get("default")

    def _get_preserved_paths(self, root_dir=None):
        """Gets list of files to preserve when exporting dataset, not including self.processed_dataset_path.

        Returns paths relative to the dataset root directory.
        """
        root_dir = root_dir if root_dir else self.processed_dataset_dir
        preserved_paths = _glob_multiple(_list_of_strings(self.config.preserve_paths), root_dir=root_dir)
        return [os.path.relpath(p, start=root_dir) for p in preserved_paths]

    def export(self, output_directory: str) -> None:
        """Exports the dataset (and any files required by it) into the specified directory."""
        self._download_and_process()
        os.makedirs(output_directory, exist_ok=True)
        shutil.copy2(self.processed_dataset_path, os.path.join(output_directory, self.processed_dataset_filename))
        preserve_paths = self._get_preserved_paths()
        for relative_path in preserve_paths:
            source = os.path.join(self.processed_dataset_dir, relative_path)
            destination = os.path.join(output_directory, relative_path)
            if os.path.isdir(source):
                shutil.copytree(source, destination, symlinks=False, dirs_exist_ok=True)
            else:
                shutil.copy2(source, destination)

    def _download_and_process(self, kaggle_username=None, kaggle_key=None):
        """Loads the dataset, downloaded and processing it if needed.

        If dataset is already processed, does nothing.
        """
        if self.state == DatasetState.NOT_LOADED:
            try:
                self.download(kaggle_username=kaggle_username, kaggle_key=kaggle_key)
            except Exception:
                logger.exception("Failed to download dataset")
        self.verify()
        if self.state == DatasetState.DOWNLOADED:
            # Extract dataset
            try:
                self.extract()
            except Exception:
                logger.exception("Failed to extract dataset")
        if self.state == DatasetState.EXTRACTED:
            # Transform dataset
            try:
                self.transform()
            except Exception:
                logger.exception("Failed to transform dataset")

    def load(self, split=False, kaggle_username=None, kaggle_key=None) -> pd.DataFrame:
        """Loads the dataset, downloaded and processing it if needed.

        Note: This method is also responsible for splitting the data, returning a single dataframe if split=False, and a
        3-tuple of train, val, test if split=True.

        :param split: (bool) splits dataset along 'split' column if present. The split column should always have values
        0: train, 1: validation, 2: test.
        """
        self._download_and_process(kaggle_username=kaggle_username, kaggle_key=kaggle_key)
        if self.state == DatasetState.TRANSFORMED:
            dataset_df = self.load_transformed_dataset()
            if split:
                return self.split(dataset_df)
            else:
                return dataset_df

    def download(self, kaggle_username=None, kaggle_key=None) -> List[str]:
        if not os.path.exists(self.raw_dataset_dir):
            os.makedirs(self.raw_dataset_dir)
        if self.is_kaggle_dataset:
            return download_kaggle_dataset(
                self.raw_dataset_dir,
                kaggle_dataset_id=self.config.kaggle_dataset_id,
                kaggle_competition=self.config.kaggle_competition,
                kaggle_username=kaggle_username,
                kaggle_key=kaggle_key,
            )
        else:
            for url, filename in zip(self.download_urls, self.download_filenames):
                downloaded_file_path = os.path.join(self.raw_dataset_dir, filename)
                with TqdmUpTo(unit="B", unit_scale=True, unit_divisor=1024, miniters=1, desc=filename) as t:
                    urllib.request.urlretrieve(url, downloaded_file_path, t.update_to)

    def verify(self) -> None:
        """Verifies checksums for dataset."""
        for filename, sha256sum in self.config.sha256.items():
            digest = _sha256_digest(os.path.join(self.raw_dataset_dir, filename))
            if digest != sha256sum:
                raise ValueError(f"Checksum mismatch for file {filename} of {self.config.name} dataset")
        if not self.config.sha256:
            logger.warning(f"No sha256 digest provided for dataset {self.config.name}, cannot verify.")
            logger.info("Contents:")
            for filename in os.listdir(self.raw_dataset_dir):
                path = os.path.join(self.raw_dataset_dir, filename)
                if not os.path.isdir(path):
                    digest = _sha256_digest(path)
                    logger.info(f"    {filename}: {digest}")

    def extract(self) -> List[str]:
        extracted_files = set()
        for download_filename in self.download_filenames:
            download_path = os.path.join(self.raw_dataset_dir, download_filename)
            if is_archive(download_path):
                extracted_files.update(extract_archive(download_path))
            # If the archive contains archives, extract those too. For example, bnp_claims_management.
            archive_contents = extracted_files.copy()
            for extracted_file in archive_contents:
                extracted_path = os.path.join(self.raw_dataset_dir, extracted_file)
                if is_archive(extracted_path):
                    try:
                        extracted_files.update(extract_archive(extracted_path))
                    except RuntimeError as e:
                        logger.warning(f"Error extracting {extracted_file}" + str(e))
        return list(extracted_files)

    def transform(self) -> None:
        data_filenames = [
            os.path.join(self.raw_dataset_dir, f) for f in os.listdir(self.raw_dataset_dir) if not is_archive(f)
        ]
        transformed_files = self.transform_files(data_filenames)
        unprocessed_dataframe = self.load_unprocessed_dataframe(transformed_files)
        transformed_dataframe = self.transform_dataframe(unprocessed_dataframe)
        self.save_processed(transformed_dataframe)

    def transform_files(self, file_paths: List[str]) -> List[str]:
        """Transform data files before loading to dataframe.

        Subclasses should override this method to process files before loading dataframe, calling the base class
        implementation after transformation if the results of transformation are needed by preserve_paths.
        """
        data_files = [p for p in file_paths if not os.path.isdir(p)]
        if not os.path.exists(self.processed_dataset_dir):
            os.makedirs(self.processed_dataset_dir)
        # Moves any preserved paths (ex. image directories) into processed directory to avoid unnecessary copy.
        for rel_path in self._get_preserved_paths(self.raw_dataset_dir):
            source_path = os.path.join(self.raw_dataset_dir, rel_path)
            dest_path = os.path.join(self.processed_dataset_dir, rel_path)
            if os.path.exists(source_path) and not os.path.exists(dest_path):
                os.replace(source_path, dest_path)
        return data_files

    def load_file_to_dataframe(self, file_path: str) -> pd.DataFrame:
        """Loads a file into a dataframe.

        Subclasses may override this method to support other input formats (json, jsonl, tsv, csv, parquet)
        """
        file_extension = os.path.splitext(file_path)[-1].lower()
        if file_extension == ".json":
            return pd.read_json(file_path)
        elif file_extension == ".jsonl":
            return pd.read_json(file_path, lines=True)
        elif file_extension == ".tsv":
            return pd.read_table(file_path)
        elif file_extension in {".csv", ".data"}:
            return pd.read_csv(file_path)
        elif file_extension in {".parquet", ".pq", ".pqt"}:
            return pd.read_parquet(file_path)
        else:
            raise ValueError(f"Unsupported dataset file type: {file_extension}")

    def load_files_to_dataframe(self, file_paths: List[str], root_dir=None) -> pd.DataFrame:
        """Loads a file or list of files and returns a dataframe.

        Subclasses may override this method to change the loader's behavior for groups of files.
        """
        if root_dir:
            file_paths = [os.path.join(root_dir, path) for path in file_paths]
        dataframes = [self.load_file_to_dataframe(path) for path in file_paths]
        if self.config.columns:
            dataframes = [df.set_axis(self.config.columns, axis=1) for df in dataframes]
        return pd.concat(dataframes, ignore_index=True)

    def load_unprocessed_dataframe(self, file_paths: List[str]) -> pd.DataFrame:
        """Load dataset files into a dataframe.

        Will use the list of data files in the dataset directory as a default if all of config's dataset_filenames,
        train_filenames, validation_filenames, test_filenames are empty.
        """
        dataset_paths = _glob_multiple(_list_of_strings(self.config.dataset_filenames), root_dir=self.raw_dataset_dir)
        train_paths = _glob_multiple(_list_of_strings(self.config.train_filenames), root_dir=self.raw_dataset_dir)
        validation_paths = _glob_multiple(
            _list_of_strings(self.config.validation_filenames), root_dir=self.raw_dataset_dir
        )
        test_paths = _glob_multiple(_list_of_strings(self.config.test_filenames), root_dir=self.raw_dataset_dir)
        dataframes = []
        if len(train_paths) > 0:
            train_df = self.load_files_to_dataframe(train_paths)
            train_df[SPLIT] = 0
            dataframes.append(train_df)
        if len(validation_paths) > 0:
            validation_df = self.load_files_to_dataframe(validation_paths)
            validation_df[SPLIT] = 1
            dataframes.append(validation_df)
        if len(test_paths) > 0:
            test_df = self.load_files_to_dataframe(test_paths)
            test_df[SPLIT] = 2
            dataframes.append(test_df)
        # If we have neither train/validation/test files nor dataset_paths in the config, use data files in root dir.
        if len(dataset_paths) == len(dataframes) == 0:
            dataset_paths = file_paths
        if len(dataset_paths) > 0:
            dataframes.append(self.load_files_to_dataframe(dataset_paths))
        return pd.concat(dataframes, ignore_index=True)

    def transform_dataframe(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        """Transforms a dataframe of the entire dataset.

        Subclasses should override this method if transformation of the dataframe is needed.
        """
        for column_name, type in self.config.column_types.items():
            dataframe[column_name] = dataframe[column_name].astype(type)
        return dataframe

    def save_processed(self, dataframe: pd.DataFrame) -> None:
        """Saves transformed dataframe as a flat file ludwig can load for training."""
        if not os.path.exists(self.processed_dataset_dir):
            os.makedirs(self.processed_dataset_dir)
        dataframe.to_parquet(self.processed_dataset_path, engine="pyarrow")

    def load_transformed_dataset(self) -> pd.DataFrame:
        """Load processed dataset into a dataframe."""
        return pd.read_parquet(self.processed_dataset_path)

    def get_mtime(self) -> float:
        """Last modified time of the processed dataset after downloading successfully."""
        return os.path.getmtime(self.processed_dataset_path)

    def split(self, dataset: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        if SPLIT in dataset:
            dataset[SPLIT] = pd.to_numeric(dataset[SPLIT])
            training_set = dataset[dataset[SPLIT] == 0].drop(columns=[SPLIT])
            val_set = dataset[dataset[SPLIT] == 1].drop(columns=[SPLIT])
            test_set = dataset[dataset[SPLIT] == 2].drop(columns=[SPLIT])
            return training_set, test_set, val_set
        else:
            raise ValueError(f"The dataset does not a '{SPLIT}' column, load with `split=False`")



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os

import pandas as pd
from scipy.io import loadmat

from ludwig.datasets.loaders.dataset_loader import DatasetLoader
from ludwig.utils.fs_utils import open_file


class SarcosLoader(DatasetLoader):
    """The Sarcos dataset.

    Details:
        The data relates to an inverse dynamics problem for a seven
        degrees-of-freedom SARCOS anthropomorphic robot arm. The
        task is to map from a 21-dimensional input space (7 joint
        positions, 7 joint velocities, 7 joint accelerations) to the
        corresponding 7 joint torques. There are 44,484 training
        examples and 4,449 test examples. The first 21 columns are
        the input variables, and the 22nd column is used as the target
        variable.

    Dataset source:
        Locally Weighted Projection RegressionL: An O(n) Algorithm for
        Incremental Real Time Learning in High Dimensional Space,
        S. Vijayakumar and S. Schaal, Proc ICML 2000.
        http://www.gaussianprocess.org/gpml/data/
    """

    def load_file_to_dataframe(self, file_path: str) -> pd.DataFrame:
        """Loads a file into a dataframe."""
        with open_file(file_path) as f:
            mat = loadmat(f)
        file_df = pd.DataFrame(mat[os.path.basename(file_path).split(".")[0]])
        return file_df

    def transform_dataframe(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        processed_df = super().transform_dataframe(dataframe)
        columns = []
        columns += [f"position_{i}" for i in range(1, 8)]
        columns += [f"velocity_{i}" for i in range(1, 8)]
        columns += [f"acceleration_{i}" for i in range(1, 8)]
        columns += [f"torque_{i}" for i in range(1, 8)]
        columns += ["split"]

        processed_df.columns = columns
        return processed_df



import numpy as np
import pandas as pd
import pytest
import torch

from ludwig.api import LudwigModel
from ludwig.constants import COLUMN, INPUT_FEATURES, OUTPUT_FEATURES
from ludwig.features.timeseries_feature import TimeseriesInputFeature
from ludwig.schema.features.timeseries_feature import TimeseriesInputFeatureConfig
from ludwig.schema.utils import load_config_with_kwargs
from tests.integration_tests.utils import number_feature, timeseries_feature

BATCH_SIZE = 2
SEQ_SIZE = 10
DEFAULT_OUTPUT_SIZE = 4


@pytest.mark.parametrize(
    "enc_encoder", ["stacked_cnn", "parallel_cnn", "stacked_parallel_cnn", "rnn", "cnnrnn", "passthrough"]
)
def test_timeseries_feature(enc_encoder):
    # synthetic time series tensor
    timeseries_tensor = torch.randn([BATCH_SIZE, SEQ_SIZE], dtype=torch.float32)

    # generate feature config
    timeseries_feature_config = timeseries_feature(
        encoder={
            "type": enc_encoder,
            "max_len": SEQ_SIZE,
            "fc_layers": [{"output_size": DEFAULT_OUTPUT_SIZE}],
            # simulated parameters determined by pre-processing
            "max_sequence_length": SEQ_SIZE,
        }
    )

    # instantiate input feature object
    timeseries_feature_config, _ = load_config_with_kwargs(TimeseriesInputFeatureConfig, timeseries_feature_config)
    timeseries_input_feature = TimeseriesInputFeature(timeseries_feature_config)

    # pass synthetic tensor through input feature
    encoder_output = timeseries_input_feature(timeseries_tensor)

    # confirm correctness of the encoder output
    assert isinstance(encoder_output, dict)
    assert "encoder_output" in encoder_output
    assert isinstance(encoder_output["encoder_output"], torch.Tensor)
    if enc_encoder == "passthrough":
        assert encoder_output["encoder_output"].shape == (BATCH_SIZE, SEQ_SIZE, 1)
    else:
        assert encoder_output["encoder_output"].shape == (BATCH_SIZE, DEFAULT_OUTPUT_SIZE)


def test_timeseries_preprocessing_with_nan():
    config = {
        "input_features": [timeseries_feature(preprocessing={"padding_value": 42})],
        "output_features": [number_feature()],
    }

    # generate synthetic data
    data = {
        config[INPUT_FEATURES][0][COLUMN]: [
            "1.53 2.3 NaN 6.4 3 ",
            "1.53 2.3 2 ",
            "1.53 NaN 3 2 ",
        ],
        config[OUTPUT_FEATURES][0][COLUMN]: [1.0, 2.0, 3.0],
    }
    df = pd.DataFrame(data)

    model = LudwigModel(config)
    ds = model.preprocess(df)
    out_df = ds.training_set.to_df()

    assert len(out_df.columns) == len(df.columns)

    expected_df = pd.DataFrame(
        [
            [np.array([1.53, 2.3, 42.0, 6.4, 3.0]), 1.0],
            [np.array([1.53, 2.3, 2.0, 42.0, 42.0]), 2.0],
            [np.array([1.53, 42.0, 3.0, 2.0, 42.0]), 3.0],
        ],
        columns=out_df.columns.to_list(),
    )

    for row1, row2 in zip(out_df.values, expected_df.values):
        assert np.allclose(row1[0], row2[0])
        assert row1[1] == row2[1]
